// HWInterface.cpp
// Direct interface to hardware: PIC, Servo Controller, etc.
#include "stdafx.h"
#include "ClientOrServer.h"
#if ( ROBOT_SERVER == 1 )	// This module used for Robot Server only

#include "MotorControl.h"
#include "DynamixelControl.h"
#include "KerrControl.h"
#include "Module.h"
#include "Globals.h"

//#if ( ROBOT_SERVER == 1 )
#include "NMEAParser.h"
//#endif

#include "HWInterface.h"
#include "HeadControl.h"
#include "LaserScanner.h"


#define PIC_SIO_BUF_LEN		 128
#define GPS_BUF_LEN			 256
#define MOTOR_RX_BUF_LEN	 256
#define SERVO_SYNC_0		0xFF	// required sync for SSC II controller
//#define SERVO_SYNC_0		0x80	// required sync for Pololu controller
#define SERVO_CONTROLLER_ID	0x01	// only one controller attached

//#if ( ROBOT_SERVER == 1 )  // Nothing in this file used for client!

#define CAMERA_UNKNOWN_POSITION 0xFFFF


__itt_string_handle* pshPicThreadReadLoop = __itt_string_handle_create("Read Loop");
__itt_string_handle* pshPicThreadWriteLoop = __itt_string_handle_create("Write Loop");

__itt_string_handle* pshReadError = __itt_string_handle_create("Read Error");
__itt_string_handle* pshPicError = __itt_string_handle_create("PIC Reported Error");
__itt_string_handle* pshHandlePicMessage = __itt_string_handle_create("Handle Pic Msg");
__itt_string_handle* pshWaitForMoreData = __itt_string_handle_create("Wait for More Data");

__itt_string_handle* pshVersionRequest = __itt_string_handle_create("VersionRequest");
__itt_string_handle* pshPicWriteFile = __itt_string_handle_create("WriteFile");
__itt_string_handle* pshWriteIndex = __itt_string_handle_create("WriteIndex = ");
__itt_string_handle* pshReadIndex = __itt_string_handle_create("ReadIndex = ");



/////////////////////////////////////////////////////////////////////////////
void HandlePicMessage(char *CmdResponseBuf, int nResponseLength)
{

	// Handles response messages from the PIC
	// ACK or AKx means things went well
	// ERR means a Serial I/O error occured
	// NAK means command not recognized or parameter error
	//TAL_Event("HandlePicMsg Top");
	//-TAL_SCOPED_TASK_NAMED("HandlePicMessage");

	CString strStatus;
	static int nReadIndex = 0; // For tracking how many reads sucessfully received

	// First byte is the CMD being acknowledged.
	int nPicCmdBeingAcknowledged = CmdResponseBuf[0];
	ROBOT_LOG( SERIAL_DBG, "\nCommand ACK for Command: %02x\n", nPicCmdBeingAcknowledged )
	CmdResponseBuf++;	// move the pointer to skip this byte now
	nResponseLength--;	// subtract one for the cmd byte

	if( nResponseLength < 3 )
	{
		strStatus.Format( "PIC COM ERR: Short Message (%d)", nResponseLength);
		ROBOT_DISPLAY( TRUE, (LPCTSTR)strStatus )
		nReadIndex = 0;
		__itt_marker(pDomainControlThread, __itt_null, pshReadError, __itt_marker_scope_task);
		return;
	}

	// We got something reasonable.  Reset the watchdog timer
	gPicWatchDogTimer = 0;

	if( 0 == strncmp(CmdResponseBuf, "ERR", 3) )
	{
		__itt_marker(pDomainControlThread, __itt_null, pshPicError, __itt_marker_scope_task);
		CmdResponseBuf[nResponseLength] = '\0';	// make sure null terminated
		#ifdef __WIN_CE_TARGET
			TCHAR szUnicodeBuffer[PIC_SIO_BUF_LEN];	// Convert to Unicode
			mbstowcs( szUnicodeBuffer, CmdResponseBuf, nResponseLength+1 ); 
			strStatus.Format( ">>>> PIC ERR: %s", szUnicodeBuffer );	// Pic will include the error
		#else
			strStatus.Format( ">>>> PIC ERR: %s", CmdResponseBuf );	// Pic will include the error
		#endif
		ROBOT_DISPLAY( TRUE, (LPCTSTR)strStatus );

	}
	else if (0 == strncmp(CmdResponseBuf, "NAK", 3) )
	{
		// Command not implemented, or bad parameter.
		__itt_marker(pDomainControlThread, __itt_null, pshPicError, __itt_marker_scope_task);
		CmdResponseBuf[nResponseLength] = '\0';	// make sure null terminated
		#ifdef __WIN_CE_TARGET
			TCHAR szUnicodeBuffer[PIC_SIO_BUF_LEN];	// Convert to Unicode
			mbstowcs( szUnicodeBuffer, CmdResponseBuf, nResponseLength+1 ); 
			strStatus.Format( ">>>> PIC: %s", szUnicodeBuffer );	// Pic will include the error
		#else
			strStatus.Format( ">>>> PIC: %s", CmdResponseBuf );	// Pic will include the error
		#endif
		ROBOT_DISPLAY( TRUE, (LPCTSTR)strStatus )
	}
	else if( 0 == strncmp(CmdResponseBuf, "AK0", 3) )
	{
		// Quick Acknowledgement with no parameters.  Nothing to do.
		ROBOT_DISPLAY( TRUE, "     PIC: AK0" )
	}
	else if( 0 == strncmp(CmdResponseBuf, "AK1", 3) )
	{
		// Standard Acknowledgement with status string.  Just print the string.
		CmdResponseBuf[nResponseLength] = '\0';	// make sure null terminated

		// remove comments if too many speed commands
//		if( (0 != strncmp(CmdResponseBuf, "AK1:SPEED", 9)) &&
//			(0 != strncmp(CmdResponseBuf, "AK1:Motor", 9)) )

		if( (0 != strncmp(CmdResponseBuf, "AK1:EYES", 8)) )	
		{
			// Don't keep showing Eye commands - happens every blink!
			strStatus.Format( "     PIC: %s", CmdResponseBuf );	// Pic will include any error
			ROBOT_DISPLAY( TRUE,  (LPCTSTR)strStatus )
		}
	}
	else if( 0 == strncmp(CmdResponseBuf, "AK2", 3) )
	{
		// Acknowledgement with Status Data.
		ROBOT_LOG( SERIAL_DBG, "PIC: AK2 Status Data\n" )	// happens every second!

		if( nResponseLength < (sizeof(PIC_STATUS_T) + 3) )
		{
			CString strStatus;
			strStatus.Format( ">>>> PIC ERR: Short Status (%d)", nResponseLength);
			ROBOT_DISPLAY( TRUE,  (LPCTSTR)strStatus )
			__itt_marker(pDomainControlThread, __itt_null, pshReadError, __itt_marker_scope_task);
			return;
		}
		memcpy( &g_RawPicStatus, &CmdResponseBuf[3], sizeof( PIC_STATUS_T ) );
		// Now throw message in the queue, to indicate that the status has been updated
		if(!g_PicFirstStatusReceived)
		{
			// Force status to be updated on the GUI, for example when client connects
			PostThreadMessage( g_dwControlThreadId, (WM_ROBOT_SENSOR_STATUS_READY), 1, 1 ); // use 1 to force status updates every time!
			g_PicFirstStatusReceived = TRUE;
		}
		else
		{
			PostThreadMessage( g_dwControlThreadId, (WM_ROBOT_SENSOR_STATUS_READY), 0, 0 );
		}
	}
	else if( 0 == strncmp(CmdResponseBuf, "AK3", 3) )
	{
		// Acknowledgement with Version Info.
		ROBOT_DISPLAY( TRUE, "     PIC AK3:Version" )

		if( nResponseLength < 4 )
		{
			__itt_marker(pDomainControlThread, __itt_null, pshReadError, __itt_marker_scope_task);
			ROBOT_DISPLAY( TRUE, ">>>> ERROR: Short Version Packet From PIC" )
			return;
		}
		int Version = CmdResponseBuf[3];
		g_PicSubSystemStatus = SUBSYSTEM_CONNECTED;	// Receipt of version is confirmation that we are connected.
		// Now throw message in the queue, to indicate that the version has been updated
		PostThreadMessage( g_dwControlThreadId, (WM_ROBOT_PIC_VERSION_READY), 0, Version );


	}
	else if( 0 == strncmp(CmdResponseBuf, "AK4", 3) )
	{
		// Acknowledgement with Radar Data.
		ROBOT_LOG( SERIAL_DBG, "PIC: AK4 Radar Data\n" )	// happens every 2 seconds!

		if( nResponseLength < (US_SCAN_MAX_SAMPLES + 4) )
		{
			__itt_marker(pDomainControlThread, __itt_null, pshReadError, __itt_marker_scope_task);
			strStatus.Format( "ERROR: Short Radar Scan (%d) From PIC", nResponseLength );
			ROBOT_DISPLAY( TRUE,  (LPCTSTR)strStatus )
			return;
		}

		if( 0 == CmdResponseBuf[3] )	// first byte indicates which sensor
		{
			// Received a scan from US Scanning Sensor 1
			memcpy( &g_BulkSensorData[US_ARRAY1][0], &CmdResponseBuf[4], US_SCAN_MAX_SAMPLES );
			PostThreadMessage( g_dwControlThreadId, (WM_ROBOT_SENSOR_DATA), 0, US_ARRAY1 );
		}
		else //if( 1 == CmdResponseBuf[3] )	
		{
			strStatus.Format( "ERROR: Scanning Sensor Number (%d) Not Implemented!", CmdResponseBuf[3] );
			ROBOT_DISPLAY( TRUE,  (LPCTSTR)strStatus )
			// Received a scan from IR Scanning Sensor 2
			//memcpy( &g_BulkSensorData[IR_ARRAY2][0], &CmdResponseBuf[4], IR_SCAN_MAX_SAMPLES );
			//PostThreadMessage( g_dwControlThreadId, (WM_ROBOT_SENSOR_DATA), 0, IR_ARRAY2 );
		}

	}
	else if( 0 == strncmp(CmdResponseBuf, "AK5", 3) )
	{
		BOOL EmptyData = TRUE;
		for( int i=3; i< nResponseLength; i++ )
		{
			if( 0 != (BYTE)CmdResponseBuf[i] )
			{
				EmptyData = FALSE;
				break;
			}
		}

		/*
		if( !EmptyData )
		{
			// DEBUG MESSAGE from PIC!
			ROBOT_LOG( TRUE,  "\nPIC DEBUG MSG: ")
			for( int i=3; i< nResponseLength; i++ )
			{
				ROBOT_LOG( TRUE,  "%02X ", (BYTE)CmdResponseBuf[i] )
			}
			ROBOT_LOG( TRUE,  "\n")
		}
		*/
	}
	else
	{
		__itt_marker(pDomainControlThread, __itt_null, pshReadError, __itt_marker_scope_task);
		strStatus.Format( "Bad PIC response: %02X %02X %02X %02X", 
			CmdResponseBuf[0], CmdResponseBuf[1], CmdResponseBuf[2], CmdResponseBuf[3]  );
		ROBOT_DISPLAY( TRUE,  (LPCTSTR)strStatus )
	}

	__itt_metadata_add(pDomainControlThread, __itt_null, pshReadIndex, __itt_metadata_s32, 1, (void*)&nReadIndex);
	nReadIndex++;

}


/////////////////////////////////////////////////////////////////////////////
DWORD WINAPI PicCommReadThreadFunc(LPVOID lpParameter)
{
	__itt_thread_set_name( "Pic Read Thread" );

	//const char PicReadCounterName[] = "PicReadCount";

	// PIC Response Parser State Machine
	#define	SIO_LOOK_FOR_SYNC_1		0
	#define	SIO_LOOK_FOR_SYNC_2		1
	#define	SIO_GET_MSG_SIZE		2
	#define	SIO_COPY_RESPONSE		3

//	DWORD fdwCommMask;
	//TCHAR szwcsBuffer[1000];
	//char szBuffer[1000];
	HANDLE hCommPort = (HANDLE)lpParameter;

	char	*pCmdResponse;
	char	SioBuffer[PIC_SIO_BUF_LEN];
	char	CmdResponseBuf[PIC_SIO_BUF_LEN];
	BOOL	bFirstPacket;
	DWORD	nSioBytesReceived;
	int		nCmdResponseLen = 0;
	int		nResponseBytesCopied;
	int		ret = 0;
	int		nSioCharNum;
	UINT	SerialParserState = SIO_LOOK_FOR_SYNC_1;
	int		DbgDotCount = 0;

	//SetCommMask (hCommPort, EV_RXCHAR | EV_BREAK );
	//Sleep(10000);	// Wait 10 seconds for the port to be ready?!?
	//Sleep(200);

	pCmdResponse = CmdResponseBuf;	// Pointer to walk the buffer, filling it up
	memset( CmdResponseBuf, '\0', PIC_SIO_BUF_LEN );	// clean out garbage chars
	memset( SioBuffer, '\0', PIC_SIO_BUF_LEN );	// clean out garbage chars
	bFirstPacket = TRUE;
	nSioBytesReceived = 0;
	nResponseBytesCopied = 0;

	while(hCommPort != INVALID_HANDLE_VALUE)
	{
		__itt_task_begin(pDomainPicThread, __itt_null, __itt_null, pshPicThreadReadLoop);

		///TAL_SCOPED_TASK_NAMED("Pic Read SIO");
		//xTAL_IncCounter(PicReadCounterName);
		// 

/*		if( SUBSYSTEM_CONNECTED != g_PicSubSystemStatus )
		{
			// WARNING! The following call will hang until a character comes in the serial port!
			// This means it will hang on shutdown unless a PIC processor is hooked up and running!
			ROBOT_LOG( TRUE, "PIC COMM Read: Waiting for Pic...\n")
			if(!WaitCommEvent (hCommPort, &fdwCommMask, 0))
			{
				// has the comm port been closed?
				if(GetLastError() != ERROR_INVALID_HANDLE)
					ReportCommError("WaitCommEvent.");
				return 0;		// terminate thread on error
			}
		}
*/
		if( SERIAL_DBG )
		{
			ROBOT_LOG( TRUE,  " [RX ");	// just show activity for each read, so we know PIC is not dead.
		}
		//SetCommMask (hCommPort, EV_RXCHAR | EV_BREAK);
		memset( SioBuffer, '\0', PIC_SIO_BUF_LEN );	// clean out garbage chars
		if( SIMULATED_SIO_HANDLE == hCommPort )
		{
			g_PicSubSystemStatus = SUBSYSTEM_CONNECTED;	// Receipt of version is confirmation that we are connected.
			__itt_task_end(pDomainPicThread);  // pshPicThreadReadLoop
			Sleep(SIMULATED_SIO_SLEEP_TIME);
			continue;
		}

		if( SERIAL_DBG )
		{
			ROBOT_LOG( TRUE,  "1,");	// just show activity for each read, so we know PIC is not dead.
		}
		if(!ReadFile(hCommPort,		// where to read from (the open comm port)
				 SioBuffer,		// where to put the character
				 (PIC_SIO_BUF_LEN-1),	// number of bytes to read
				 &nSioBytesReceived,	// how many bytes were actually read
				 NULL))				// overlapped I/O not supported
		{

			ROBOT_DISPLAY( TRUE, "ERROR Reading PIC Comm port\n" )
			PurgeComm(hCommPort, PURGE_RXABORT|PURGE_RXCLEAR);
			//PurgeComm(hCommPort, PURGE_RXABORT|PURGE_TXABORT|PURGE_RXCLEAR|PURGE_TXCLEAR);
			DWORD dwCommError = 0;
			ClearCommError(hCommPort, &dwCommError, NULL);
			ReportCommError("PIC SIO Read Error", dwCommError );

			__itt_marker(pDomainControlThread, __itt_null, pshReadError, __itt_marker_scope_task);
			__itt_task_end(pDomainPicThread);  // pshPicThreadReadLoop
			//ROBOT_ASSERT(0);	// TODO REMOVE THIS - looking for cause of PIC errors DWS
			continue; // CHANGED TO TEST PIC RELIABILITY! DWS return 0;		// terminate thread on first error
		}

		if( SERIAL_DBG )
		{
			ROBOT_LOG( TRUE,  "2,");	// just show activity for each read, so we know PIC is not dead.
		}
		if( 0 == nSioBytesReceived )
		{
			//ROBOT_LOG( SERIAL_DBG, "PIC COMM Read - no data ready.\n" )
			if( SERIAL_DBG )
			{
				ROBOT_LOG( TRUE,  "0] ");	// just show activity for each read, so we know PIC is not dead.
			}

			__itt_task_end(pDomainPicThread);  // pshPicThreadReadLoop
			if( SUBSYSTEM_CONNECTED != g_PicSubSystemStatus )
			{
				Sleep(100);
			}
			else
			{
				Sleep(50);
			}
			continue;
		}		
		if( SERIAL_DBG )
		{
			ROBOT_LOG( TRUE,  "3 RX CHAR! ,");	// just show activity for each read, so we know PIC is not dead.
		}

		///////////////////////// Buffer Parse Loop /////////////////////////////////
		char* pParser = SioBuffer;	// get a pointer to the buffer

		{
			///TAL_SCOPED_TASK_NAMED("Parse Buffer From PIC");

/*			if( SERIAL_DBG )
			{
				ROBOT_LOG( TRUE,  " [RX] ");	// just show activity for each read, so we know PIC is not dead.
				if( DbgDotCount++ > 30 )
				{
					ROBOT_LOG( TRUE,  "\n" );
					DbgDotCount = 0;
				}
			}	
*/
		if( SERIAL_DBG )
		{
			ROBOT_LOG( TRUE,  "4,");	// just show activity for each read, so we know PIC is not dead.
		}
			int nMsgBytesReceived = 0;	
			for (nSioCharNum=0; nSioCharNum< (int)nSioBytesReceived; nSioCharNum++)
			{
				switch( SerialParserState )
				{
					case SIO_LOOK_FOR_SYNC_1:
					{
						// Look for Sync Frame (2 "Z" chars in a row)
						if( 'Z' == *pParser++ )
						{
							// got first Sync char.  
							SerialParserState = SIO_LOOK_FOR_SYNC_2;
						}
					}
					break;

					case SIO_LOOK_FOR_SYNC_2:
					{
						if( 'Z' == *pParser++ )
						{
							// Second Sync found
							SerialParserState = SIO_GET_MSG_SIZE;
						}
						else
						{
							// error, bail!
							SerialParserState = SIO_LOOK_FOR_SYNC_1;
						}
					}
					break;

					case SIO_GET_MSG_SIZE:
					{
						// Next byte is the size of the packet
						nCmdResponseLen = (int)*pParser++;	// Length the PIC said the message should be!
						if( nCmdResponseLen > RESPONSE_STRING_LEN+1 )	// Longest Message allowed
						{
							ROBOT_DISPLAY( TRUE, "ERROR: Invalid Pic Response Length!" )
							SerialParserState = SIO_LOOK_FOR_SYNC_1;
							__itt_marker(pDomainControlThread, __itt_null, pshReadError, __itt_marker_scope_task);
						}
						else if( 0 == nCmdResponseLen )
						{
							ROBOT_DISPLAY( TRUE, "ERROR: Pic Response Length Zero!" )
							SerialParserState = SIO_LOOK_FOR_SYNC_1;
							__itt_marker(pDomainControlThread, __itt_null, pshReadError, __itt_marker_scope_task);
						}
						else
						{
							bFirstPacket = FALSE;
							SerialParserState = SIO_COPY_RESPONSE;
						}
					}
					break;

					case SIO_COPY_RESPONSE:
					{

						// Save response, without the size byte or sync bytes
						nMsgBytesReceived = nSioBytesReceived - nSioCharNum; // Subtract out leading sync and length bytes
						if( (nResponseBytesCopied + nMsgBytesReceived) >= nCmdResponseLen )	// nResponseBytesCopied holds bytes recieved previously
						{
							// We received the full message now, with possible extra 
							// bytes for the next message in the buffer too!
							// Copy just the remaining bytes for THIS message
							int BytesToCopy = nCmdResponseLen - nResponseBytesCopied;
							memcpy( pCmdResponse, pParser, BytesToCopy );

							nSioCharNum += (BytesToCopy-1);	// Move counter up for the number of bytes copied out of the SIO buffer, minus the next i++
							pParser += BytesToCopy;		// Same for the input pointer
							nResponseBytesCopied += BytesToCopy;

							if( nResponseBytesCopied != nCmdResponseLen )
							{
								ROBOT_LOG( TRUE, "ERROR: BytesCopied != CmdResponseLen!\n")
								__itt_marker(pDomainControlThread, __itt_null, pshReadError, __itt_marker_scope_task);
							}

							// Sucessfully read the full response!
							CmdResponseBuf[nCmdResponseLen+1] = '\0';	// Null terminate
	//						ROBOT_LOG( TRUE, "Start:HandlePicMessage\n")
							
							//Sleep(5);
							{
								///TAL_Event("PIC Msg Received!");
								///TAL_SCOPED_TASK_NAMED("Call  HandlePicMessage");
								__itt_task_begin(pDomainPicThread, __itt_null, __itt_null, pshHandlePicMessage);
								HandlePicMessage(CmdResponseBuf, nCmdResponseLen);
								__itt_task_end(pDomainPicThread);
							}
	//						ROBOT_LOG( TRUE, "End:HandlePicMessage\n")

							// Initialize for new message
							nResponseBytesCopied = 0;
							pCmdResponse = CmdResponseBuf;
							memset( CmdResponseBuf, '\0', PIC_SIO_BUF_LEN );	// clean out garbage chars
							
							SerialParserState = SIO_LOOK_FOR_SYNC_1;
						}
						else
						{
							// The full response is not in the buffer, so 
							// just copy what we've received so far
							memcpy( pCmdResponse, pParser, nMsgBytesReceived ); 
							pCmdResponse = pCmdResponse + nMsgBytesReceived;
							nResponseBytesCopied += nMsgBytesReceived;
							nSioCharNum += (nMsgBytesReceived-1);	// for/next loop should exit after this
							__itt_marker(pDomainControlThread, __itt_null, pshWaitForMoreData, __itt_marker_scope_task);
						}
					}
					break;

					default:
					{
						ROBOT_DISPLAY( TRUE, "ERROR!  Bad STATE in PicCommReadThreadFunc" )
						SerialParserState = SIO_LOOK_FOR_SYNC_1;
					}

				}	// Switch state

			}	// For each char in incoming buffer
		}	// TAL ParseData

		if( SERIAL_DBG )
		{
			ROBOT_LOG( TRUE,  "5] ");	// just show activity for each read, so we know PIC is not dead.
		}

		__itt_task_end(pDomainPicThread);  // pshPicThreadReadLoop
		Sleep(0); // force thread switch after each read

	}	// while port open

	ROBOT_LOG( TRUE, "PIC COMM Thread exiting.\n" )
	return 0;
}

DWORD WINAPI PicCommWriteThreadFunc(LPVOID lpParameter)
{
	// Writes a command to the PIC and waits for an ACKnowlegement
	// Messages are sent by SendHardwareCmd calls
	// This thread is accessed via g_dwPicCommWriteThreadId
	IGNORE_UNUSED_PARAM(lpParameter);
	__itt_thread_set_name( "Pic Write Thread" );

	//const char PicWriteCounterName[] = "PicWriteCount";

	MSG		msg;
	char	CmdBuf[16];
	DWORD	dwBytesWritten;
	BYTE	Request;
	BYTE	Index;
	BYTE	Value;
	BYTE	Option1;
	BYTE	Option2;
//	WORD	TempWord;
	CString strStatus;
//	BYTE	Checksum;

	DWORD	nSioResponseTime = 0;
	DWORD	nSioStartTime = 0;
	UINT	PicRetryCount = 0;
	int		DbgDotCount = 0;

	static int nWriteIndex = 0;


	// Process message loop for this thread
	while( 0 != GetMessage( &msg, NULL, WM_ROBOT_MESSAGE_BASE, (WM_ROBOT_MESSAGE_BASE+HW_MAX_MESSAGE) ) )
		// OK to filter out Windows messages, since we don't use WM_QUIT
	{
		__itt_task_begin(pDomainPicThread, __itt_null, __itt_null, pshPicThreadWriteLoop);

		Request = (BYTE)(msg.message - WM_ROBOT_MESSAGE_BASE);
		if( WM_ROBOT_THREAD_EXIT_CMD == msg.message ) 
		{
			break; // Quit command received!
		}

		if ( INVALID_HANDLE_VALUE == g_hPicCommPort )
		{
			// Usually, we should get a WM_QUIT before the COMM port closes, 
			// but if not, go ahead and exit...
			ROBOT_LOG( TRUE, "\nPIC COMM WRITE: Port INVALID HANDLE \n" )
			__itt_task_end(pDomainPicThread);  // pshPicThreadWriteLoop
			return 0;	
		}

		if( (msg.message < WM_ROBOT_MESSAGE_BASE) || ((msg.message > (WM_ROBOT_MESSAGE_BASE+HW_MAX_MESSAGE) ) ) )
		{
			// Ignore any messages not intended for user (where do these come from?)
			ROBOT_LOG( TRUE, "PicCommWriteThread: message out of Range, WM_ROBOT_MESSAGE_BASE = 0x%08X, Msg = 0x%08lX, wParam = 0x%08lX, lParam = 0x%08lX\n", 
				WM_ROBOT_MESSAGE_BASE, msg.message, msg.wParam, msg.lParam )
			__itt_task_end(pDomainPicThread);  // pshPicThreadWriteLoop
			continue;
		}

		if ( (g_DownloadingFirmware) && (Request != HW_RESET_CPU) )
		{
			ROBOT_LOG( TRUE, ">>> >>> DOWNLOADING FIRMWARE TO PIC!  CMD IGNORED!!!  <<< <<<<\n")
			__itt_task_end(pDomainPicThread);  // pshPicThreadWriteLoop
			continue;
		}

		if( Request == HW_RESET_CPU ) 
		{
			ROBOT_LOG( TRUE, ">>> Sending RESET_CPU to PIC!\n")
		}

		// Get ready to send the message to the PIC
		Index = (BYTE)(msg.wParam);
		Value = (BYTE)(msg.lParam);
		Option1 = 0;
		Option2 = 0;

		// TODO: CheckSum = Request+Index+Value+Option1+Option2;	// Rolls over and drops bits
		// Copy command and parameters into buffer
		CmdBuf[0] = (char)(BYTE)SIO_SYNC_0;
		CmdBuf[1] = (char)(BYTE)SIO_SYNC_1;
		CmdBuf[2] = (char)Request;
		CmdBuf[3] =	(char)Index;
		CmdBuf[4] =	(char)Value;
		CmdBuf[5] =	(char)Option1;
		CmdBuf[6] =	(char)Option2;
		CmdBuf[7] = (char)(BYTE)CMD_TERM_CHAR;	// Check to make sure no bytes were dropped

		if( HW_GET_VERSION == Request )
		{
			///TAL_Event("Pic Requesting Version");
			//ROBOT_DISPLAY( TRUE, "\nPIC COMM WRITE: Requesting Version..." )
			__itt_marker(pDomainControlThread, __itt_null, pshVersionRequest, __itt_marker_scope_task);
			nWriteIndex = 0;
		}
		else
		{
			__itt_metadata_add(pDomainControlThread, __itt_null, pshWriteIndex, __itt_metadata_s32, 1, (void*)&nWriteIndex);
			nWriteIndex++;
		}

		BOOL bSucess = 0;

		ROBOT_LOG( SERIAL_DBG, "PIC CMD: Sending cmd %02X to Pic.\n", Request )
		// Don't retry or wait for status, version or Joystick commands
		if( SIMULATED_SIO_HANDLE == g_hPicCommPort )
		{
			__itt_task_end(pDomainPicThread);  // pshPicThreadWriteLoop
			Sleep(SIMULATED_SIO_SLEEP_TIME);
			continue;
		}


		//////////////////////////////////////////////////////////////////////////
		// Write to the COM Port
		__itt_task_begin(pDomainPicThread, __itt_null, __itt_null, pshPicWriteFile);
		bSucess = WriteFile(g_hPicCommPort,	// where to write to (the open comm port)
				CmdBuf,				// what to write
				SERIAL_CMD_SIZE,	// number of bytes to be written to port
				&dwBytesWritten,	// number of bytes that were actually written
				NULL);				// overlapped I/O not supported
		__itt_task_end(pDomainPicThread);
		//////////////////////////////////////////////////////////////////////////

		if( !bSucess )
		{
			ROBOT_DISPLAY( TRUE, "ERROR Writing PIC Comm port\n" )
			PurgeComm(g_hPicCommPort, PURGE_RXABORT|PURGE_RXCLEAR);
			//PurgeComm(hCommPort, PURGE_RXABORT|PURGE_TXABORT|PURGE_RXCLEAR|PURGE_TXCLEAR);
			DWORD dwCommError = 0;
			ClearCommError(g_hPicCommPort, &dwCommError, NULL);
			ReportCommError("PIC SIO Write Error", dwCommError );

			__itt_task_end(pDomainPicThread);  // pshPicThreadWriteLoop
			continue;
		}
		else
		{
			if(SERIAL_CMD_SIZE != dwBytesWritten)
			{
				strStatus.Format( "PIC SERIAL ERROR Bytes Written = %d!\n", dwBytesWritten );
				ROBOT_DISPLAY( TRUE,  (LPCTSTR)strStatus )
				__itt_task_end(pDomainPicThread);  // pshPicThreadWriteLoop
				continue;
			}

			ROBOT_LOG( SERIAL_DBG, "Sent %d bytes to Pic.\n", dwBytesWritten )
			if( SERIAL_DBG )
			{
				ROBOT_LOG( TRUE,  " [TX] ");	// just show activity for each write, so we know PIC is not dead.
				if( DbgDotCount++ > 30 )
				{
					ROBOT_LOG( TRUE,  "\n" );
					DbgDotCount = 0;
				}
			}

			if( Request == HW_RESET_CPU ) 
			{
				ROBOT_LOG( TRUE, ">>> PIC RESET_CPU command SENT!\n")
			}
		}

		__itt_task_end(pDomainPicThread);  // pshPicThreadReadLoop
		Sleep(0); // force thread switch after each command
	}
	ROBOT_LOG( TRUE, "\nPIC COMM WRITE: Received WM_QUIT\n")
	return 0;
}

////////////////////////////////////////////////////////////////////////////////////////
__itt_string_handle* pshGPSReadLoop = __itt_string_handle_create("GPS Read Loop");
DWORD WINAPI GPSCommReadThreadFunc(LPVOID lpParameter)
{
	HANDLE hCommPort = (HANDLE)lpParameter;
	__itt_thread_set_name( "GPS Read Thread" );

	BYTE	GPSBuffer[GPS_BUF_LEN];
	DWORD dwBytesReceived;

	// Create the GPS Parser object
	CNMEAParser NMEAParser;
	NMEAParser.m_dGSAVDOP = 22;


	if( SIMULATED_SIO_HANDLE != g_hPicCommPort )
	{
		SetCommMask (g_hGPSCommPort, EV_RXCHAR | EV_BREAK);
	}
	Sleep(4000);	// Allow other startups, and for the port to be ready

	while(g_hGPSCommPort != INVALID_HANDLE_VALUE)
	{
		__itt_task_begin(pDomainGPSThread, __itt_null, __itt_null, pshGPSReadLoop);

		if( SIMULATED_SIO_HANDLE == g_hPicCommPort )
		{
			__itt_task_end(pDomainGPSThread);  // pshGPSReadLoop
			Sleep(100); //Use instead of SIMULATED_SIO_SLEEP_TIME for GPS
			continue;
		}		

		//SetCommMask (g_hGPSCommPort, EV_RXCHAR | EV_BREAK);
		dwBytesReceived = 0;
		GPSBuffer[0] = '\0';
		// Read ANSI characters
		if(!ReadFile(g_hGPSCommPort,		// where to read from (the open comm port)
				 GPSBuffer,			// where to put the character
				 (GPS_BUF_LEN-1),	// number of bytes to read
				 &dwBytesReceived,		// how many bytes were actually read
				 NULL))				// no overlapped I/O
		{

			ROBOT_DISPLAY( TRUE, "ERROR Reading GPS Comm port\n" )
			PurgeComm(g_hGPSCommPort, PURGE_RXABORT|PURGE_RXCLEAR);
			//PurgeComm(hCommPort, PURGE_RXABORT|PURGE_TXABORT|PURGE_RXCLEAR|PURGE_TXCLEAR);
			DWORD dwCommError = 0;
			ClearCommError(g_hGPSCommPort, &dwCommError, NULL);
			ReportCommError("GPS SIO Read Error", dwCommError );

			__itt_task_end(pDomainGPSThread);  // pshGPSReadLoop
			return 0;		// terminate thread on error
		}
		if( 0 == dwBytesReceived )
		{
			// Just keep looking for data from the device
			__itt_task_end(pDomainGPSThread);  // pshGPSReadLoop
			Sleep(100);
			continue;
		}
		else
		{
			// Got some GPS data.  Parse it.
			if( NMEAParser.ParseBuffer(GPSBuffer, dwBytesReceived) )
			{
				// Data ready - Notify the robot control engine!
				// ROBOT_LOG( TRUE,"HWInterface - Got Data!\n")
				g_GPSSubSystemStatus = SUBSYSTEM_CONNECTED;

				// Now throw message in the queue, to indicate that the GPS data has been updated
				PostThreadMessage( g_dwControlThreadId, (WM_ROBOT_GPS_DATA_READY), 0, 0 );
			}
		}

		__itt_task_end(pDomainGPSThread);  // pshGPSReadLoop
		Sleep(100);

	}	// while port open


	ROBOT_LOG( TRUE, "GPS COMM Read Thread exiting.\n")
	return 0;
}



////////////////////////////////////////////////////////////////////////////////////////
__itt_string_handle* pshPololuWriteLoop = __itt_string_handle_create("Pololu Write Loop");

DWORD WINAPI ServoCommWriteThreadFunc(LPVOID lpParameter)
{
	// Writes a command to the Pololu Servo controller
	// Messages are sent by SendHardwareCmd calls
	IGNORE_UNUSED_PARAM(lpParameter);
	__itt_thread_set_name( "Pololu Servo Write Thread" );

	MSG		msg;
	char	CmdBuf[16];
	DWORD	dwBytesWritten;
	BYTE	Request;
	BYTE	Index;
	BYTE	Value;
	BYTE	Option1;
	BYTE	Option2;
//	WORD	TempWord;
	CString strStatus;
	int		DbgDotCount = 0;
	BYTE ServoNumber = 0;
	BYTE Position = 0;

	Sleep(3000);

	// Process message loop for this thread
	while( 0 != GetMessage( &msg, NULL, WM_ROBOT_MESSAGE_BASE, (WM_ROBOT_MESSAGE_BASE+HW_MAX_MESSAGE) ) )
		// OK to filter out Windows messages, since we don't use WM_QUIT
	{
		__itt_task_begin(pDomainPololuServoThread, __itt_null, __itt_null, pshPololuWriteLoop);

		Request = (BYTE)(msg.message - WM_ROBOT_MESSAGE_BASE);

		if( WM_ROBOT_THREAD_EXIT_CMD == Request )
		{
			break; // Quit command received!
		}

		if( (msg.message < WM_ROBOT_MESSAGE_BASE) || ((msg.message > (WM_ROBOT_MESSAGE_BASE+HW_MAX_MESSAGE) ) ) )
		{
			// Ignore any messages not intended for user (where do these come from?)
			ROBOT_LOG( TRUE, "PicCommWriteThread: message out of Range, WM_ROBOT_MESSAGE_BASE = 0x%08X, Msg = 0x%08lX, wParam = 0x%08lX, lParam = 0x%08lX\n", 
				WM_ROBOT_MESSAGE_BASE, msg.message, msg.wParam, msg.lParam )
			__itt_task_end(pDomainPololuServoThread);  // pshPololuWriteLoop
			continue;
		}
		// 

		// Get ready to send the message to the Servo controller

		// TODO!  TEST THIS!
		// WARNING! WARNING!  CHANGED THIS FROM UNPACKING BYTES TO USING WPARAM AND LPARAM DIRECTLY!
		// THIS MAY BREAK PIC COMMUNICATION - NEED TO TEST!!!
//		TempWord = (WORD)(msg.lParam);
//		Index = HIBYTE(msg.wParam);
//		Value = LOBYTE(msg.wParam);
//		Option1 = HIBYTE(TempWord);
//		Option2 = LOBYTE(TempWord);
		Index = (BYTE)(msg.wParam);
		Value = (BYTE)(msg.lParam);
		Option1 = 0;
		Option2 = 0;


		// Format Servo command
		// We don't use the Pololu mode for now...
		// case HW_INITIALIZE_SERVO:
		// case HW_SET_SERVO_ZERO_POS:
		// CmdBuf[1] = SERVO_CONTROLLER_ID;


		switch( Request )
		{
			case HW_SET_MOTOR_STOP:
			{
				ServoNumber = SERVO_SPEED;
				Position = SPEED_SERVO_CENTER;	// Mini SSC II mode: 0x00 to 0xFE
				break;
			}
/*			No longer used!
			case HW_SET_MOVE_DISTANCE:
			{
				ROBOT_DISPLAY( TRUE, "ERROR! ServoCommWriteThread: HW_SET_MOVE_DISTANCE not supported!\n" )
				break;
			}
*/
			case HW_SET_TURN:
			{
				ServoNumber = SERVO_TURN;
				Position = Index;	// Mini SSC II mode: 0x00 to 0xFE
				break;
			}
			case HW_SET_SPEED:
			{
				ServoNumber = SERVO_SPEED;
				Position = Index;	// Mini SSC II mode: 0x00 to 0xFE
				break;
			}
			case HW_SET_GEAR:
			{
				ServoNumber = SERVO_GEAR;
				Position = Value;	// Mini SSC II mode: 0x00 to 0xFE
				ROBOT_LOG( TRUE,"Extern Servo SET GEAR: %d", Position)
				break;
			}

			case HW_SET_CAMERA_PAN_ABS:
			{
				ServoNumber = SERVO_CAMERA_PAN + SERVO_MODE_180; // See Pololu doc for SSEII mode
				Position = Index;	// Mini SSC II mode: 0x00 to 0xFE
				break;
			}
			case HW_SET_CAMERA_TILT_ABS:
			{
				ServoNumber = SERVO_CAMERA_TILT;
				Position = (255-Index);	// Tilt Servo is Inverted
				break;
			}
			case HW_SET_CAMERA_PAN_TILT_ABS:
			{
				// This function can only hand commands that control one servo at a time!
				ROBOT_DISPLAY( TRUE, "ERROR! ServoCommWriteThread: HW_SET_CAMERA_PAN_TILT_ABS not supported!\n" )
				ROBOT_ASSERT(0);
				break;
			}

			case HW_SET_SPEED_AND_TURN:
			{
				// This function can only hand commands that control one servo at a time!
				ROBOT_DISPLAY( TRUE, "ERROR! ServoCommWriteThread: HW_SET_SPEED_AND_TURN not supported!\n" )
				break;
			}
			default:
			{
				CString StrText;
				StrText.Format( "ERROR! ServoCommWriteThread: Unknown Cmd:%02X \n", Request);
				ROBOT_DISPLAY( TRUE, (StrText) )
			}

		}

		//TAL_Event("Sending Cmd");
		//-TAL_SCOPED_TASK_NAMED("Servo Write SIO");

		CmdBuf[0] = (char)(BYTE)SERVO_SYNC_0;
		CmdBuf[1] =	(char)ServoNumber;
		CmdBuf[2] =	(char)Position;

		ROBOT_LOG( SERIAL_DBG, "Servo CMD: Sending cmd %02X to Servo controller.\n", Request )
		if( SIMULATED_SIO_HANDLE == g_hPicCommPort )
		{
			__itt_task_end(pDomainPololuServoThread);  // pshPololuWriteLoop
			Sleep(SIMULATED_SIO_SLEEP_TIME);
			continue;
		}
		if(!WriteFile(g_hServoCommPort,	// where to write to (the open comm port)
				CmdBuf,				// what to write
				3,					// number of bytes to be written to port
				&dwBytesWritten,	// number of bytes that were actually written
				NULL))				// overlapped I/O not supported			
		{
			ROBOT_DISPLAY( TRUE, "SERIAL ERROR Sending Servo data!\n" )
			__itt_task_end(pDomainPololuServoThread);  // pshPololuWriteLoop
			continue;
		}
		else
		{
			ROBOT_LOG( SERIAL_DBG, "Sent %d bytes to Servo.\n", dwBytesWritten )
			if( SERIAL_DBG )
			{
				ROBOT_LOG( TRUE,  "Tx");	// just show activity for each write, so we know Servo is not dead.
				if( DbgDotCount++ > 30 )
				{
					ROBOT_LOG( TRUE,  "\n" );
					DbgDotCount = 0;
				}
			}

		}
		__itt_task_end(pDomainPololuServoThread);  // pshPololuWriteLoop
		Sleep(0); // force thread switch after each command
	}
	ROBOT_LOG( TRUE, "SERVO COMM WRITE: Received WM_QUIT\n")
	return 0;
}

void SendCameraCmd(char *CameraBytes, int nByteCount)
{
	DWORD dwBytesWritten;

	ASSERT(0); // Serial Camera not enabled!
	if( SIMULATED_SIO_HANDLE == g_hPicCommPort )
	{
		Sleep(SIMULATED_SIO_SLEEP_TIME);
		return;
	}
	if(!WriteFile(g_hCameraCommPort,	// where to write to (the open comm port)
			CameraBytes,			// what to write
			nByteCount,		// number of bytes to be written to port
			&dwBytesWritten,	// number of bytes that were actually written
			NULL))				// overlapped I/O not supported			
	{
		ROBOT_LOG( TRUE,"Serial Error Sending Camera data.\n")
	}
	else
	{
		ROBOT_LOG( SERIAL_DBG, "Sent %d bytes to Camera.\n", dwBytesWritten )
	}
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////
#if CAMERA_CONTROL_TYPE == SONY_SERIAL_CAMERA

DWORD WINAPI CameraCommWriteThreadFunc(LPVOID lpParameter)
{
	// Writes a command to the Sony Camera serial interface
	// Messages are sent by SendHardwareCmd calls
	__itt_thread_set_name( "Sony Camera Write Thread" );


	MSG		msg;
	BYTE	Request;
	CString strStatus;
	int		DbgDotCount = 0;
	BYTE CameraNumber = 0;
	BYTE Position = 0;
	static int CurrentPanPos = 0;
	static int CurrentTiltPos = 0;
	static int AbsPanSpeed = 20;
	static int AbsTiltSpeed = 16;


	// Process message loop for this thread
	while( 0 != GetMessage( &msg, NULL, WM_ROBOT_MESSAGE_BASE, (WM_ROBOT_MESSAGE_BASE+HW_MAX_MESSAGE) ) )
		// OK to filter out Windows messages, since we don't use WM_QUIT
	{
		Request = msg.message - WM_ROBOT_MESSAGE_BASE;

		if( WM_ROBOT_THREAD_EXIT_CMD == msg.message ) 
		{
			break; // Quit command received!
		}

		if( INVALID_HANDLE_VALUE == g_hCameraCommPort )
		{
			continue; // Camera is not connected.
		}

		if( (msg.message < WM_ROBOT_MESSAGE_BASE) || ((msg.message > (WM_ROBOT_MESSAGE_BASE+HW_MAX_MESSAGE) ) ) )
		{
			// Ignore any messages not intended for user (where do these come from?)
			ROBOT_LOG( TRUE, "PicCommWriteThread: message out of Range, WM_ROBOT_MESSAGE_BASE = 0x%08X, Msg = 0x%08lX, wParam = 0x%08lX, lParam = 0x%08lX\n", 
				WM_ROBOT_MESSAGE_BASE, msg.message, msg.wParam, msg.lParam )
			continue;
		}

		int wParam = msg.wParam;
		int lParam = msg.lParam;

		// Note that lParam is only used for Abs Pan/Tilt, and HW_SET_CAMERA_MODE commands

		switch( Request )
		{		

			case HW_SET_CAMERA_MODE:
			{
				// Put Camera in a specific mode
				// wParam = Mode, lParam = enable/disable
				if( CAMERA_MODE_DISPLAY_ENABLE == wParam)
				{
					// Set Camera Display mode

					char CameraBytes[6];
					CameraBytes[0] = (char)0x81;	// Header for Camera #1
					CameraBytes[1] = (char)0x01;	// Control Command
					CameraBytes[2] = (char)0x06;	// 0606 = Display control command 
					CameraBytes[3] = (char)0x06;	//  
									// On/Off Command will be inserted here
					CameraBytes[5] = (char)0xFF;	// Terminator

					if( 1 == lParam )
					{
						CameraBytes[4] = (char)0x02;	// Display On
					}
					else
					{
						CameraBytes[4] = (char)0x03;	// Display Off
					}
					SendCameraCmd(CameraBytes, 6);			
				}
				else if( CAMERA_MODE_TRACK_FACE == wParam )
				{
					ROBOT_LOG( TRUE,"NOT IMPLEMENTED - CAMERA_MODE_TRACK_FACE\n")
					ROBOT_ASSERT(0);	
				}
				else
				{
					ROBOT_ASSERT(0);	// ERROR BAD MODE!
				}

				break;
			}



			case HW_SET_CAMERA_ABS_PAN_TILT_SPEED:
			{
				// Just save the values for later.
				// SONY: Pan speed 0-18h (24), Tilt speed 0-14h (20)
				if( (UINT)(wParam) > SERVO_SPEED_MAX ) wParam = SERVO_SPEED_MAX;		// Can't be negative, UINT
				AbsPanSpeed = wParam + 2;
				AbsTiltSpeed = wParam - 2;
				if( AbsTiltSpeed < 0 ) AbsTiltSpeed = 0;	// Limit zero
				break;
			}

			case HW_SET_CAMERA_PAN_ABS:
			{
				// Limit values to max camera can handle
				if( wParam > SONY_CAMERA_MAX_RIGHT ) wParam = SONY_CAMERA_MAX_RIGHT;
				if( wParam < SONY_CAMERA_MAX_LEFT ) wParam = SONY_CAMERA_MAX_LEFT;

				CurrentPanPos = wParam;	// Save value in case an absolute tilt command comes in
				DWORD PanPos = (DWORD)wParam;
				DWORD TiltPos = (DWORD)CurrentTiltPos;

				char CameraAbsBytes[15];
				CameraAbsBytes[0] =  (char)0x81;	// Header for Camera #1
				CameraAbsBytes[1] =  (char)0x01;	// Control Command
				CameraAbsBytes[2] =  (char)0x06;	// 06 02 = Pan/Tilt Absolute command
				CameraAbsBytes[3] =  (char)0x02;
				CameraAbsBytes[4] =  (char)AbsPanSpeed;	// Pan speed (0-18h)
				CameraAbsBytes[5] =  (char)AbsTiltSpeed;// Tilt speed (0-14h)
				CameraAbsBytes[6] =  (char)0x00;	// Pan -----------------------------------
				CameraAbsBytes[7] =  (char)0x00;	// Range: 0xFC90 to 0x0370 = -880  <-> +880
				CameraAbsBytes[8] =  (char)0x00;	// every high nibble is zero: 0Y 0Y 0Y 0Y
				CameraAbsBytes[9] =  (char)0x00;	// 
				CameraAbsBytes[10] = (char)0x00;	// Tilt -----------------------------------
				CameraAbsBytes[11] = (char)0x00;	// Range: 0xFED4 to 0x012C = -300 <-> +300
				CameraAbsBytes[12] = (char)0x00;	// every high nibble is zero: 0Y 0Y 0Y 0Y
				CameraAbsBytes[13] = (char)0x00;	// 
				CameraAbsBytes[14] = (char)0xFF;	// Terminator

				// Loop and shift, copying one Nibble at a time into the buffer!
				// Note data is Sparsely Populated, and reverse-indian.  13AF ==> 0F 0A 03 01
				for(int Nibble=0; Nibble < 4; Nibble++ )
				{
					CameraAbsBytes[(9-Nibble)] = (PanPos & 0x000F);
					PanPos =  PanPos >> 4;

					CameraAbsBytes[(13-Nibble)] = (TiltPos & 0x000F);
					TiltPos = TiltPos >> 4;
				}
				
				SendCameraCmd(CameraAbsBytes, 15);
				ROBOT_LOG( TRUE,"Debug Camera - sent Pan Command, Position = %04x\n", (DWORD)wParam )
				break;
			}
			case HW_SET_CAMERA_TILT_ABS:
			{
				// Limit values to max camera can handle
				if( wParam > SONY_CAMERA_MAX_UP ) wParam = SONY_CAMERA_MAX_UP;
				if( wParam < SONY_CAMERA_MAX_DOWN ) wParam = SONY_CAMERA_MAX_DOWN;

				CurrentTiltPos = wParam;	// Save value in case an absolute tilt command comes in
				DWORD TiltPos = (DWORD)wParam;
				DWORD PanPos = (DWORD)CurrentPanPos;

				char CameraAbsBytes[15];
				CameraAbsBytes[0] =  (char)0x81;	// Header for Camera #1
				CameraAbsBytes[1] =  (char)0x01;	// Control Command
				CameraAbsBytes[2] =  (char)0x06;	// 06 02 = Pan/Tilt Absolute command
				CameraAbsBytes[3] =  (char)0x02;
				CameraAbsBytes[4] =  (char)AbsPanSpeed;	// Pan speed (0-18h)
				CameraAbsBytes[5] =  (char)AbsTiltSpeed;// Tilt speed (0-14h)
				CameraAbsBytes[6] =  (char)0x00;	// Pan -----------------------------------
				CameraAbsBytes[7] =  (char)0x00;	// Range: 0xFC90 to 0x0370 = -880  <-> +880
				CameraAbsBytes[8] =  (char)0x00;	// every high nibble is zero: 0Y 0Y 0Y 0Y
				CameraAbsBytes[9] =  (char)0x00;	// 
				CameraAbsBytes[10] = (char)0x00;	// Tilt -----------------------------------
				CameraAbsBytes[11] = (char)0x00;	// Range: 0xFED4 to 0x012C = -300 <-> +300
				CameraAbsBytes[12] = (char)0x00;	// every high nibble is zero: 0Y 0Y 0Y 0Y
				CameraAbsBytes[13] = (char)0x00;	// 
				CameraAbsBytes[14] = (char)0xFF;	// Terminator

				// Loop and shift, copying one Nibble at a time into the buffer!
				for(int Nibble=0; Nibble < 4; Nibble++ )
				{
					CameraAbsBytes[(9-Nibble)] = (PanPos & 0x000F);
					PanPos =  PanPos >> 4;

					CameraAbsBytes[(13-Nibble)] = (TiltPos & 0x000F);
					TiltPos = TiltPos >> 4;
				}
				
				SendCameraCmd(CameraAbsBytes, 15);
				ROBOT_LOG( TRUE,"Debug Camera - sent Tilt Command, Position = %04x\n", (DWORD)wParam )
				break;
			}
			case HW_SET_CAMERA_PAN_TILT_ABS:
			{
				// wParam = Pan, lParam = tilt
				// Limit values to max camera can handle
				if( wParam > SONY_CAMERA_MAX_RIGHT ) wParam = SONY_CAMERA_MAX_RIGHT;
				else if( wParam < SONY_CAMERA_MAX_LEFT ) wParam = SONY_CAMERA_MAX_LEFT;

				if( lParam > SONY_CAMERA_MAX_UP ) lParam = SONY_CAMERA_MAX_UP;
				else if( lParam < SONY_CAMERA_MAX_DOWN ) lParam = SONY_CAMERA_MAX_DOWN;

				CurrentPanPos = wParam;		// Save value in case an absolute pan command comes in
				CurrentTiltPos = lParam;	// Save value in case an absolute tilt command comes in
				DWORD PanPos = (DWORD)wParam;
				DWORD TiltPos = (DWORD)lParam;

				char CameraAbsBytes[15];
				CameraAbsBytes[0] =  (char)0x81;	// Header for Camera #1
				CameraAbsBytes[1] =  (char)0x01;	// Control Command
				CameraAbsBytes[2] =  (char)0x06;	// 06 02 = Pan/Tilt Absolute command
				CameraAbsBytes[3] =  (char)0x02;
				CameraAbsBytes[4] =  (char)AbsPanSpeed;	// Pan speed (0-18h)
				CameraAbsBytes[5] =  (char)AbsTiltSpeed;// Tilt speed (0-14h)
				CameraAbsBytes[6] =  (char)0x00;	// Pan -----------------------------------
				CameraAbsBytes[7] =  (char)0x00;	// Range: 0xFC90 to 0x0370 = -880  <-> +880
				CameraAbsBytes[8] =  (char)0x00;	// every high nibble is zero: 0Y 0Y 0Y 0Y
				CameraAbsBytes[9] =  (char)0x00;	// 
				CameraAbsBytes[10] = (char)0x00;	// Tilt -----------------------------------
				CameraAbsBytes[11] = (char)0x00;	// Range: 0xFED4 to 0x012C = -300 <-> +300
				CameraAbsBytes[12] = (char)0x00;	// every high nibble is zero: 0Y 0Y 0Y 0Y
				CameraAbsBytes[13] = (char)0x00;	// 
				CameraAbsBytes[14] = (char)0xFF;	// Terminator

				// Loop and shift, copying one Nibble at a time into the buffer!
				for(int Nibble=0; Nibble < 4; Nibble++ )
				{
					CameraAbsBytes[(9-Nibble)] = (PanPos & 0x000F);
					PanPos =  PanPos >> 4;

					CameraAbsBytes[(13-Nibble)] = (TiltPos & 0x000F);
					TiltPos = TiltPos >> 4;
				}
				
				SendCameraCmd(CameraAbsBytes, 15);
//				ROBOT_LOG( TRUE,"Debug Camera - sent Pan+Tilt Command, Pan = %04d, Tilt = %04d\n", wParam, lParam )
				break;
			}

			case HW_CAMERA_INITIALIZE:
			{
				// Send Camera Initialization string
				char CameraBytes[4];
				CameraBytes[0] = (char)0x88; 
				CameraBytes[1] = (char)0x30; 
				CameraBytes[2] = (char)0x01; 
				CameraBytes[3] = (char)0xFF;

				for( int i=0; i<10; i++)
				{
					SendCameraCmd(CameraBytes, 4);		//TODO - is this really needed?
					Sleep(500);
				}
				CurrentPanPos = 0;
				CurrentTiltPos = 0;
				break;
			}
			case HW_SET_CAMERA_ZOOM:
			{
				int ZoomCmd = wParam;
				// TODO: int ZoomSpeed = lParam;
				char CameraBytes[6];
				CameraBytes[0] = (char)0x81;	// Header for Camera #1
				CameraBytes[1] = (char)0x01;	// Control Command
				CameraBytes[2] = (char)0x04;	// 04 = Zoom command
				CameraBytes[3] = (char)0x07;	// 07 = continuous zoom
								// Zoom Command will be inserted here
				CameraBytes[5] = (char)0xFF;	// Terminator

				if( CAMERA_ZOOM_STOP == ZoomCmd )
				{
						CameraBytes[4] = (char)0x00;	// Zoom Stop
						SendCameraCmd(CameraBytes, 6);
				}
				else if( CAMERA_ZOOM_IN == ZoomCmd )
				{
						CameraBytes[4] = (char)0x24;	// Zoom In TODO - Use ZoomSpeed (2-7) for low nibble
						SendCameraCmd(CameraBytes, 6);
				}
				else if( CAMERA_ZOOM_OUT == ZoomCmd )
				{
						CameraBytes[4] = (char)0x34;	// Zoom Out TODO - Use ZoomSpeed (2-7) for low nibble
						SendCameraCmd(CameraBytes, 6);
				}
				else
				{
						ROBOT_LOG( TRUE, "ERROR! Undefined WM_ROBOT_USER_CAMERA_ZOOM_CMD Option!\n")
				}
				break;
			}
///////////////////////////////////////////////////////////////////////////////////

			case HW_SET_CAMERA_ZOOM_ABS:
			{
				int ZoomLevel = wParam;		// Standard Range: 0 - 16
				// standard Zoom range is 0 to 16.  Map this to Sony range of 0 to 03FF (1023)
				ZoomLevel = ZoomLevel * 64;				
				// Limit values to max camera can handle
				if( ZoomLevel > SONY_CAMERA_MAX_ZOOM_ABS ) ZoomLevel = SONY_CAMERA_MAX_ZOOM_ABS;
				if( ZoomLevel < SONY_CAMERA_MIN_ZOOM_ABS ) ZoomLevel = SONY_CAMERA_MIN_ZOOM_ABS;
				ROBOT_LOG( TRUE,"Debug Camera - sent Abs Zoom Command, Position = %04x\n", (DWORD)ZoomLevel )

				// TODO: int ZoomSpeed = lParam;
				char CameraBytes[9];
				CameraBytes[0] = (char)0x81;	// Header for Camera #1
				CameraBytes[1] = (char)0x01;	// Control Command
				CameraBytes[2] = (char)0x04;	// 04 = Zoom command
				CameraBytes[3] = (char)0x47;	// 47 = Absolute Zoom
				CameraBytes[4] = (char)0x00;	// Absolute Zoom will be inserted here
				CameraBytes[5] = (char)0x00;	// Absolute Zoom will be inserted here
				CameraBytes[6] = (char)0x00;	// Absolute Zoom will be inserted here
				CameraBytes[7] = (char)0x00;	// Absolute Zoom will be inserted here
				CameraBytes[8] = (char)0xFF;	// Terminator

				// Loop and shift, copying one Nibble at a time into the buffer!
				// Note data is Sparsely Populated, and reverse-indian.  13AF ==> 0F 0A 03 01
				for(int Nibble=0; Nibble < 4; Nibble++ )
				{
					CameraBytes[(7-Nibble)] = (ZoomLevel & 0x000F);
					ZoomLevel =  ZoomLevel >> 4;
				}

				SendCameraCmd(CameraBytes, 9);
				break;
			}
///////////////////////////////////////////////////////////////////////////////////

			case HW_SET_CAMERA_POWER:
			{
				int PowerCmd = lParam;
				// wParam not used
				char CameraBytes[6];
				CameraBytes[0] = (char)0x81;	// Header for Camera #1
				CameraBytes[1] = (char)0x01;	// Control Command
				CameraBytes[2] = (char)0x04;	// 04 00 = Power command
				CameraBytes[3] = (char)0x00;
								// Command will be inserted here
				CameraBytes[5] = (char)0xFF;	// Terminator

				CurrentPanPos = 0;
				CurrentTiltPos = 0;

				if( POWER_ON == PowerCmd )
				{
					CameraBytes[4] = (char)0x02;	// Off
					SendCameraCmd(CameraBytes, 6);
				}
				else if( POWER_OFF == PowerCmd )
				{
					CameraBytes[4] = (char)0x03;	// On
					SendCameraCmd(CameraBytes, 6);
				}
				else if( CAMERA_RESET == PowerCmd )
				{
					CameraBytes[0] = (char)0x88;	// Header for all Cameras (broadcast message)
					CameraBytes[1] = (char)0x01;	// Control Command
					CameraBytes[2] = (char)0x00;	// 00 01 = Clear command
					CameraBytes[3] = (char)0x01;	
					CameraBytes[4] = (char)0xFF;	// Terminator
					SendCameraCmd(CameraBytes, 5);
				}
				else	
				{
						ROBOT_LOG( TRUE, "ERROR! Undefined WM_ROBOT_CAMERA_POWER_CMD Option!\n")
				}

				break;
			}
			case HW_SET_CAMERA_PAN_TILT:
			{
				int Direction = wParam;	// Pan/Tilt direction
				BYTE nPanTiltSpeed = lParam;
				char CameraBytes[9];
				CameraBytes[0] = (char)0x81;	// Header for Camera #1
				CameraBytes[1] = (char)0x01;	// Control Command
				CameraBytes[2] = (char)0x06;	// 06 01 = Pan/Tilt command
				CameraBytes[3] = (char)0x01;
				CameraBytes[4] = (char)nPanTiltSpeed;	// Pan speed
				CameraBytes[5] = (char)nPanTiltSpeed;	// Tilt speed
								// Pan Command will be inserted here
				CameraBytes[8] = (char)0xFF;	// Terminator
				CurrentPanPos = CAMERA_UNKNOWN_POSITION;	// scan does not leave us in a known position
				CurrentTiltPos = CAMERA_UNKNOWN_POSITION;

				switch( Direction )  // Which direction to Pan/Tilt
				{

					case CAMERA_PAN_STOP:
					{
						CameraBytes[6] = (char)0x03;	// No Pan
						CameraBytes[7] = (char)0x03;	// No Tilt
						SendCameraCmd(CameraBytes, 9);
						break;
					}

					// Pan or Tilt Commands

					case CAMERA_PAN_UP:
					{
						CameraBytes[6] = (char)0x03;	// No Pan
						CameraBytes[7] = (char)0x01;	// Up
						SendCameraCmd(CameraBytes, 9);
						break;
					}

					case CAMERA_PAN_DOWN:
					{
						CameraBytes[6] = (char)0x03;	// No Pan
						CameraBytes[7] = (char)0x02;	// Down
						SendCameraCmd(CameraBytes, 9);
						break;
					}

					case CAMERA_PAN_LEFT:
					{
						CameraBytes[6] = (char)0x01;	// Pan Left
						CameraBytes[7] = (char)0x03;	// No Tilt
						SendCameraCmd(CameraBytes, 9);
						break;
					}

					case CAMERA_PAN_RIGHT:
					{
						CameraBytes[6] = (char)0x02;	// Pan Right
						CameraBytes[7] = (char)0x03;	// No Tilt
						SendCameraCmd(CameraBytes, 9);
						break;
					}

				// Combo Pan/Tilt commands

					case CAMERA_PAN_UP_LEFT:
					{
						CameraBytes[6] = (char)0x01;	// Pan
						CameraBytes[7] = (char)0x01;	// Tilt
						SendCameraCmd(CameraBytes, 9);
						break;
					}

					case CAMERA_PAN_UP_RIGHT:
					{
						CameraBytes[6] = (char)0x02;	// Pan
						CameraBytes[7] = (char)0x01;	// Tilt
						SendCameraCmd(CameraBytes, 9);
						break;
					}

					case CAMERA_PAN_DOWN_LEFT:
					{
						CameraBytes[6] = (char)0x01;	// Pan
						CameraBytes[7] = (char)0x02;	// Tilt
						SendCameraCmd(CameraBytes, 9);
						break;
					}

					case CAMERA_PAN_DOWN_RIGHT:
					{
						CameraBytes[6] = (char)0x02;	// Pan
						CameraBytes[7] = (char)0x02;	// Tilt
						SendCameraCmd(CameraBytes, 9);
						break;
					}

					case CAMERA_PAN_ABS_CENTER:
					{
						char CameraAbsBytes[15];
						CameraAbsBytes[0] =  (char)0x81;	// Header for Camera #1
						CameraAbsBytes[1] =  (char)0x01;	// Control Command
						CameraAbsBytes[2] =  (char)0x06;	// 06 02 = Pan/Tilt Absolute command
						CameraAbsBytes[3] =  (char)0x02;
						CameraAbsBytes[4] =  (char)0x11;	// Pan speed (0-18h)
						CameraAbsBytes[5] =  (char)0x0D;	// Tilt speed (0-14h)
						CameraAbsBytes[6] =  (char)0x00;	// Pan Center 0000
						CameraAbsBytes[7] =  (char)0x00;	// Range: 0x0370 - 0xFC90
						CameraAbsBytes[8] =  (char)0x00;	// 
						CameraAbsBytes[9] =  (char)0x00;	// 
						CameraAbsBytes[10] = (char)0x00;	// Tilt Center 0000
						CameraAbsBytes[11] = (char)0x00;	// Range: 0x012C - 0xFED4
						CameraAbsBytes[12] = (char)0x00;	// 
						CameraAbsBytes[13] = (char)0x00;	// 
						CameraAbsBytes[14] = (char)0xFF;	// Terminator
						SendCameraCmd(CameraAbsBytes, 15);
						CurrentPanPos = 0;
						CurrentTiltPos = 0;
						break;
					}

					default:
					{
						ROBOT_LOG( TRUE, "ERROR! Invalid Camera Pan Direction, wParam = 0x%08lX\n", wParam )
					}
					break;
				}
			}	
			break;

			default:
			{
				CString StrText;
				StrText.Format( "ERROR! CameraCommWriteThread: Unknown Cmd:%02X \n", Request);
				ROBOT_DISPLAY( TRUE, (StrText)
			}

		}
	}
	return 0;
}
#endif	// CAMERA_CONTROL_TYPE == SONY_SERIAL_CAMERA



////////////////////////////////////////////////////////////////////////////////////////////
//                             WHEEL MOTOR CONTROL
////////////////////////////////////////////////////////////////////////////////////////////
#if MOTOR_CONTROL_TYPE == ER1_MOTOR_CONTROL
	// For ER1 Pilot motor controller
	DWORD WINAPI MotorCommThreadFunc(LPVOID lpParameter)
	{
		// Writes a command to the ER1 Pilot Motor controller and
		// Reads a response to each command
		// Messages are sent to this thread by SendHardwareCmd calls
		__itt_thread_set_name( "ER1 Motor Thread" );

		MSG		msg;
	//	char	CmdBuf[16];
	//	DWORD	dwBytesWritten;
		UINT	Request;
		UINT	Index;
		UINT	Value;
		UINT	Option1;
		UINT	Option2;
	//	WORD	TempWord;
		CString strStatus;
		int		DbgDotCount = 0;
		BYTE MotorNumber = 0;
		BYTE Position = 0;


		CPilotControl PilotControl;	// PMD Pilot Command handler Class
		Sleep(1000);			// Allow other initialization to complete
		PilotControl.Init();	// Send Initialization commands

		// Process message loop for this thread


	/*		if ( INVALID_HANDLE_VALUE == g_hMotorCommPort )
			{
				// Usually, we should get a WM_QUIT before the COMM port closes, 
				// but if not, go ahead and exit...
				ROBOT_LOG( TRUE, "PILOT COMM WRITE: Port INVALID HANDLE \n")
				return 0;	
			}
	*/


			// TODO!  TEST THIS!
			// WARNING! WARNING!  CHANGED THIS FROM UNPACKING BYTES TO USING WPARAM AND LPARAM DIRECTLY!
			// THIS MAY BREAK PIC COMMUNICATION - NEED TO TEST!!!
	//		TempWord = (WORD)(msg.lParam);
	//		Index = HIBYTE(msg.wParam);
	//		Value = LOBYTE(msg.wParam);
	//		Option1 = HIBYTE(TempWord);
	//		Option2 = LOBYTE(TempWord);
			Index = (UINT)(msg.wParam);
			Value = (UINT)(msg.lParam);
			Option1 = 0;
			Option2 = 0;


			PilotControl.HandlePilotCommand( Request, Index, Value, Option1, Option2 );

			Sleep(0); // force thread switch after each command
		}
		ROBOT_LOG( TRUE, "PILOT MOTOR COMM THREAD: Received WM_QUIT\n")
		return 0;
	}


////////////////////////////////////////////////////////////////////////////////////////////
#elif MOTOR_CONTROL_TYPE == POLOLU_TREX_MOTOR_CONTROL
	__itt_string_handle* pshMotorThreadLoop = __itt_string_handle_create("Motor Loop");

	// For Pololu TReX motor controller
	DWORD WINAPI MotorCommThreadFunc(LPVOID lpParameter)
	{
		///TAL_SetThreadName("Motor Thread");
		// Writes a command to the Pololu TReX Motor controller and
		// Reads a responses as needed
		// Messages are sent to this thread by SendHardwareCmd calls
		IGNORE_UNUSED_PARAM(lpParameter);
		__itt_thread_set_name( "Wheel Motor Thread" );

		MSG		msg;
	//	char	CmdBuf[16];
	//	DWORD	dwBytesWritten;
		UINT	Request;
		UINT	Index;
		UINT	Value;
	//	UINT	Option1;
	//	UINT	Option2;
	//	WORD	TempWord;
		CString strStatus;
		int		DbgDotCount = 0;
		BYTE MotorNumber = 0;
		BYTE Position = 0;


		CTrexControl TrexControl;	// Pololu Trex Command handler Class
		Sleep(2000);				// Allow other initialization to complete
		TrexControl.Init();			// Send Initialization commands

		// Process message loop for this thread
		while( 0 != GetMessage( &msg, NULL, WM_ROBOT_MESSAGE_BASE, (WM_ROBOT_MESSAGE_BASE+HW_MAX_MESSAGE) ) )
			// OK to filter out Windows messages, since we don't use WM_QUIT
		{
			//TAL_Event("Sending Cmd");
			///TAL_SCOPED_TASK_NAMED("Motor CMD Loop");
			__itt_task_begin(pDomainMotorThread, __itt_null, __itt_null, pshMotorThreadLoop);


			if( WM_ROBOT_THREAD_EXIT_CMD == msg.message ) 
			{
				break; // Quit command received!
			}

	/*		if ( INVALID_HANDLE_VALUE == g_hMotorCommPort )
			{
				// Usually, we should get a WM_QUIT before the COMM port closes, 
				// but if not, go ahead and exit...
				ROBOT_LOG( TRUE, "TReX COMM WRITE: Port INVALID HANDLE \n")
				return 0;	
			}
	*/
			if( (msg.message < WM_ROBOT_MESSAGE_BASE) || ((msg.message > (WM_ROBOT_MESSAGE_BASE+HW_MAX_MESSAGE) ) ) )
			{
				// Ignore any messages not intended for user (where do these come from?)
				ROBOT_LOG( TRUE, "MotorCommThreadFunc: message out of Range, WM_ROBOT_MESSAGE_BASE = 0x%08X, Msg = 0x%08lX, wParam = 0x%08lX, lParam = 0x%08lX\n", 
					WM_ROBOT_MESSAGE_BASE, msg.message, msg.wParam, msg.lParam )
				__itt_task_end(pDomainMotorThread);  // pshMotorThreadLoop
				continue;
			}

			Request = msg.message - WM_ROBOT_MESSAGE_BASE;
			Index = (UINT)(msg.wParam);
			Value = (UINT)(msg.lParam);

			{
				///TAL_SCOPED_TASK_NAMED("Trex HandleCmd");
				//TAL_Event("Handle Motor Command");
				TrexControl.HandleCommand( Request, Index, Value );
			}

			__itt_task_end(pDomainMotorThread);  // pshMotorThreadLoop
			Sleep(0);
		}

		ROBOT_LOG( TRUE, "TREX MOTOR COMM THREAD: Received WM_ROBOT_THREAD_EXIT_CMD\n")
		return 0;
	}

////////////////////////////////////////////////////////////////////////////////////////////
#elif MOTOR_CONTROL_TYPE == IROBOT_MOTOR_CONTROL
	__itt_string_handle* pshMotorThreadLoop = __itt_string_handle_create("Motor Loop");

	// For iRobot Create motor controller
	DWORD WINAPI MotorCommThreadFunc(LPVOID lpParameter)	// g_hMotorCommPort
	{
		///TAL_SetThreadName("Motor Thread");
		// Writes a command to the iRobot base controller
		// Status message are read by a separate thread
		// Messages are sent to this thread by SendHardwareCmd calls
		IGNORE_UNUSED_PARAM(lpParameter);
		__itt_thread_set_name( "Wheel Motor Thread" );

		MSG		msg;
		UINT	Request;
		UINT	Index;
		UINT	Value;
		CString strStatus;
		//int		DbgDotCount = 0;
		//BYTE	MotorNumber = 0;
		//BYTE	Position = 0;

		CiRobotControl iRobotControl;	// iRobot Command handler Class
		//Sleep(100);					// Allow other initialization to complete
		ROBOT_LOG( TRUE,"\n\n************* INIT IROBOT *****************\n\n")
		iRobotControl.Init();			// Send Initialization commands - starts the status read stream

		// Process message loop for this thread
		while( 0 != GetMessage( &msg, NULL, WM_ROBOT_MESSAGE_BASE, (WM_ROBOT_MESSAGE_BASE+HW_MAX_MESSAGE) ) )
			// OK to filter out Windows messages, since we don't use WM_QUIT
		{
			//TAL_Event("Sending Cmd");
			///TAL_SCOPED_TASK_NAMED("Motor CMD Loop");
			__itt_task_begin(pDomainMotorThread, __itt_null, __itt_null, pshMotorThreadLoop);


			if( WM_ROBOT_THREAD_EXIT_CMD == msg.message ) 
			{
				break; // Quit command received!
			}

	/*		if ( INVALID_HANDLE_VALUE == g_hMotorCommPort )
			{
				// Usually, we should get a WM_QUIT before the COMM port closes, 
				// but if not, go ahead and exit...
				ROBOT_LOG( TRUE, "TReX COMM WRITE: Port INVALID HANDLE \n")
				return 0;	
			}
	*/
			if( (msg.message < WM_ROBOT_MESSAGE_BASE) || ((msg.message > (WM_ROBOT_MESSAGE_BASE+HW_MAX_MESSAGE) ) ) )
			{
				// Ignore any messages not intended for user (where do these come from?)
				ROBOT_LOG( TRUE, "MotorCommThreadFunc: message out of Range, WM_ROBOT_MESSAGE_BASE = 0x%08X, Msg = 0x%08lX, wParam = 0x%08lX, lParam = 0x%08lX\n", 
					WM_ROBOT_MESSAGE_BASE, msg.message, msg.wParam, msg.lParam )
				__itt_task_end(pDomainMotorThread);  // pshMotorThreadLoop
				continue;
			}

			Request = msg.message - WM_ROBOT_MESSAGE_BASE;
			Index = (UINT)(msg.wParam);
			Value = (UINT)(msg.lParam);

			{
				///TAL_SCOPED_TASK_NAMED("Trex HandleCmd");
				//TAL_Event("Handle Motor Command");
				iRobotControl.HandleCommand( Request, Index, Value );
			}

			__itt_task_end(pDomainMotorThread);  // pshMotorThreadLoop
			Sleep(0);
		}

		ROBOT_LOG( TRUE, "TREX MOTOR COMM THREAD: Received WM_ROBOT_THREAD_EXIT_CMD\n")
		return 0;
	}

////////////////////////////////////////////////////////////////////////////////////////
__itt_string_handle* pshiRobotReadThreadLoop = __itt_string_handle_create("iRobot Read Loop");
__itt_string_handle* pshiRobotParseBuffer = __itt_string_handle_create("iRobot Parse");
DWORD WINAPI iRobotCommReadThreadFunc(LPVOID lpParameter) // g_hMotorCommPort
{
	IGNORE_UNUSED_PARAM (lpParameter);
	__itt_thread_set_name( "iRobot Read Thread" );

//	HANDLE hCommPort = (HANDLE)lpParameter;	// Use Global instead
//	DWORD	fdwCommMask;
	DWORD	dwSIOBytesReceived = 0;
	char *iRobotSIOBuf = new char[IROBOT_READ_BUF_SIZE];
	memset( iRobotSIOBuf,0,IROBOT_READ_BUF_SIZE );

	// Create the iRobot Parser object
	CiRobotParser *piRobotParser = new CiRobotParser;

	DWORD dwBytesReceived;

	if( SIMULATED_SIO_HANDLE != g_hPicCommPort )
	{
		SetCommMask( g_hMotorCommPort, EV_RXCHAR | EV_BREAK );
	}
	Sleep(500);	// Alow other initialization to occur

	while(g_hMotorCommPort != INVALID_HANDLE_VALUE)
	{
		///TAL_Event("Laser Read Loop Top");
		///TAL_SCOPED_TASK_NAMED("Laser Read Loop");
		__itt_task_begin(pDomainMotorThread, __itt_null, __itt_null, pshiRobotReadThreadLoop);

		dwBytesReceived = 0;
		iRobotSIOBuf[0] = '\0';
		//memset( iRobotSIOBuf,0,IROBOT_READ_BUF_SIZE );	// DEBUG

		BOOL bSucess = 0;
		{
			if( SIMULATED_SIO_HANDLE == g_hMotorCommPort )
			{
				Sleep(SIMULATED_SIO_SLEEP_TIME);
				continue;
			}

			bSucess = ReadFile(g_hMotorCommPort,		// where to read from (the open comm port)
					 iRobotSIOBuf,						// where to put the character
					 (IROBOT_READ_BUF_SIZE-1),			// number of bytes to read
					 &dwSIOBytesReceived,				// how many bytes were actually read
					 NULL);								// no overlapped I/O
		}
		if( !bSucess ) 
		{
			PurgeComm(g_hMotorCommPort, PURGE_RXABORT|PURGE_RXCLEAR);
			//PurgeComm(hCommPort, PURGE_RXABORT|PURGE_TXABORT|PURGE_RXCLEAR|PURGE_TXCLEAR);
			DWORD dwCommError = 0;
			ClearCommError(g_hMotorCommPort, &dwCommError, NULL);
			ReportCommError("Reading iRobot Comm port.", dwCommError );
		}
		else if( 0 != dwSIOBytesReceived )
		{
			//ROBOT_LOG( TRUE,"iROBOT Read Thread: %d bytes received\n", dwSIOBytesReceived )

			// Got some data.  Parse it.
			__itt_task_begin(pDomainMotorThread, __itt_null, __itt_null, pshiRobotParseBuffer);
			if( piRobotParser->ParseBuffer(iRobotSIOBuf, dwSIOBytesReceived) )
			{
				// Recieved binary status
				g_MotorSubSystemStatus = SUBSYSTEM_CONNECTED;
			}
			else
			{
				g_MotorSubSystemStatus = SUBSYSTEM_FAILED;
			}
			__itt_task_end(pDomainMotorThread);  // pshiRobotParseBuffer
		}

		__itt_task_end(pDomainMotorThread);  // pshLaserReadThreadLoop
		Sleep(30); // TODO!!! don't spin in a tight loop!  iRobot sends data every 15ms - IS THIS OK?  DOES ReadFile sleep?
	}	// while port open


	ROBOT_LOG( TRUE,"iROBOT COMM Read Thread exiting.\n")
	SAFE_DELETE( piRobotParser );
	SAFE_DELETE( iRobotSIOBuf );
	return 0;
}





#else
	#error BAD SENSOR_CONFIG_TYPE!  
#endif  // MOTOR_CONTROL_TYPE



////////////////////////////////////////////////////////////////////////////////////////////
// For Smart Servos
__itt_string_handle* pshSmartServoThreadLoop = __itt_string_handle_create("SmartServo Loop");
__itt_string_handle* pshDynaHandleCmd = __itt_string_handle_create("Dyna HandleCmd");
__itt_string_handle* pshKerrHandleCmd = __itt_string_handle_create("Kerr HandleCmd");

DWORD WINAPI SmartServoCommThreadFunc(LPVOID lpParameter)	// g_dwSmartServoCommThreadId
{
	// Writes a command to the Dynamixel and Kerr (Shoulder) Servos
	// and reads a response to each command
	// Messages are sent to this thread by SendHardwareCmd calls
	IGNORE_UNUSED_PARAM(lpParameter);
	__itt_thread_set_name( "Smart Servo Thread" );

	MSG		msg;
	UINT	Request;
	UINT	Index;
	UINT	Value;
	UINT	Option1;
	UINT	Option2;
	CString strStatus;
	int		DbgDotCount = 0;
	BYTE	MotorNumber = 0;
	BYTE	Position = 0;

	CDynaControl DynaControl;	// Dynamixel Control handler
#if( TURTLE != ROBOT_TYPE )
	CKerrControl KerrControl;	// Kerr Control handler
	Sleep(100);					// Allow other initialization to complete

	DynaControl.Init();			// Send Initialization commands
	Sleep(100);					// Allow other initialization to complete
	KerrControl.Init();			// Send Initialization commands

#else
	Sleep(100);					// Allow other initialization to complete
	DynaControl.Init();			// Send Initialization commands

#endif

	// Process message loop for this thread
	while( 0 != GetMessage( &msg, NULL, WM_ROBOT_MESSAGE_BASE, (WM_ROBOT_MESSAGE_BASE+HW_MAX_MESSAGE) ) )
		// OK to filter out Windows messages, since we don't use WM_QUIT
	{
		///TAL_SCOPED_TASK_NAMED("Smart Servo Loop");
		//TAL_Event("Sending Cmd");
		__itt_task_begin(pDomainSmartServoThread, __itt_null, __itt_null, pshSmartServoThreadLoop);

		Request = msg.message - WM_ROBOT_MESSAGE_BASE;
		// ROBOT_LOG( TRUE, "DEBUG SmartServoCommThreadFunc: msg.message = %04X \n", msg.message)

		if( WM_ROBOT_THREAD_EXIT_CMD == msg.message ) 
		{
			break; // Quit command received!
		}

		// Usually, we should get a WM_QUIT before the COMM port closes, but if not, go ahead and exit...
		if ( INVALID_HANDLE_VALUE == g_hDynaServoCommPort_AX12 )
		{
			ROBOT_LOG( TRUE, "SmartServoCommThreadFunc: AX12 Port INVALID HANDLE g_hSmartServoCommPort_AX12\n")
			__itt_task_end(pDomainSmartServoThread);  // pshSmartServoThreadLoop
			return 0;	
		}
#if( LOKI == ROBOT_TYPE )
		if ( INVALID_HANDLE_VALUE == g_hDynaServoCommPort_RX64 )
		{
			ROBOT_LOG( TRUE, "SmartServoCommThreadFunc: RX64 Port INVALID HANDLE g_hSmartServoCommPort_RX64\n")
			__itt_task_end(pDomainSmartServoThread);  // pshSmartServoThreadLoop
			return 0;	
		}
		if ( INVALID_HANDLE_VALUE == g_hKerrServoCommPort )
		{
			ROBOT_LOG( TRUE, "SmartServoCommThreadFunc: KERR Port INVALID HANDLE g_hKerrServoCommPort\n")
			__itt_task_end(pDomainSmartServoThread);  // pshSmartServoThreadLoop
			return 0;	
		}
#endif
		if( (msg.message < WM_ROBOT_MESSAGE_BASE) || ((msg.message > (WM_ROBOT_MESSAGE_BASE+HW_MAX_MESSAGE) ) ) )
		{
			// Ignore any messages not intended for user (where do these come from?)
			ROBOT_LOG( TRUE, "SmartServoCommThreadFunc: message out of Range, WM_ROBOT_MESSAGE_BASE = 0x%08X, Msg = 0x%08lX, wParam = 0x%08lX, lParam = 0x%08lX\n", 
				WM_ROBOT_MESSAGE_BASE, msg.message, msg.wParam, msg.lParam )
			__itt_task_end(pDomainSmartServoThread);  // pshSmartServoThreadLoop
			continue;
		}

		Index = (UINT)(msg.wParam);
		Value = (UINT)(msg.lParam);
		Option1 = 0;
		Option2 = 0;


		if( HW_SET_CAMERA_PAN_TILT_ABS == Request )
		{
			ROBOT_ASSERT(0);
			// Not supported.
		}

		///TAL_SCOPED_TASK_NAMED("Handle SmartServoCMD");
		//TAL_Event("HandleSmartServoCommand");
#if( LOKI == ROBOT_TYPE )
		__itt_task_begin(pDomainSmartServoThread, __itt_null, __itt_null, pshKerrHandleCmd);
		KerrControl.HandleCommand( Request, Index, Value, Option1, Option2 );
		__itt_task_end(pDomainSmartServoThread);  // pshKerrHandleCmd
#endif
		__itt_task_begin(pDomainSmartServoThread, __itt_null, __itt_null, pshDynaHandleCmd);
		DynaControl.HandleCommand( Request, Index, Value, Option1, Option2 );
		__itt_task_end(pDomainSmartServoThread);  // pshDynaHandleCmd

		__itt_task_end(pDomainSmartServoThread);  // pshSmartServoThreadLoop
		Sleep(0); // force thread switch after each command
	}
	__itt_task_end(pDomainSmartServoThread);  // pshSmartServoThreadLoop
	ROBOT_LOG( TRUE, "Smart Servo Control Thread: Received WM_ROBOT_THREAD_EXIT_CMD\n")
	CloseCommPort( "Dynamixel AX12" , g_hDynaServoCommPort_AX12 );
	CloseCommPort( "Dynamixel RX64" , g_hDynaServoCommPort_RX64 );
	CloseCommPort( "Kerr" , g_hKerrServoCommPort );
	ROBOT_LOG( TRUE, "SmartServoCommThreadFunc: Smart Servo COMM ports closed.  Exiting thread.\n")

	return 0;
}



////////////////////////////////////////////////////////////////////////////////////////////
// For Hokuyo URG-04LX-UG01 Laser Scanner
__itt_string_handle* pshLaserWriteThreadLoop = __itt_string_handle_create("Laser Write Loop");
__itt_string_handle* pshLaserHandleCommand = __itt_string_handle_create("Handle Command");
DWORD WINAPI LaserScannerCommWriteThreadFunc(LPVOID lpParameter)	// g_dwLaserScannerCommWriteThreadId, g_hLaserScannerCommPort
{
	///TAL_SetThreadName("Laser Write Thread");
	// Writes a command to the Laser Scanner
	// Messages are sent to this thread by SendHardwareCmd calls
	IGNORE_UNUSED_PARAM(lpParameter);
	__itt_thread_set_name( "Laser Write Thread" );

	MSG		msg;
//	char	CmdBuf[16];
//	DWORD	dwBytesWritten;
	UINT	Request;
	CString strStatus;

	CLaserScannerCommand LaserScannerCommand;	// Laser Scanner Control handler Class
	Sleep(5000);								// Allow other initialization to complete
	Sleep(5000);								
	LaserScannerCommand.Init();					// Send Initialization commands (reset?)

	// Process message loop for this thread
	while( 0 != GetMessage( &msg, NULL, WM_ROBOT_MESSAGE_BASE, (WM_ROBOT_MESSAGE_BASE+HW_MAX_MESSAGE) ) )
		// OK to filter out Windows messages, since we don't use WM_QUIT
	{
		///TAL_SCOPED_TASK_NAMED("Laser Write Loop");
		//TAL_Event("Sending Cmd");
		__itt_task_begin(pDomainLaserThread, __itt_null, __itt_null, pshLaserWriteThreadLoop);


		Request = msg.message - WM_ROBOT_MESSAGE_BASE;

//		ROBOT_LOG( TRUE, "DEBUG LaserScannerCommWriteThreadFunc: msg.message = %04X \n", msg.message)

		if( WM_ROBOT_THREAD_EXIT_CMD == msg.message ) 
		{
			break; // Quit command received!
		}

		if ( INVALID_HANDLE_VALUE == g_hLaserScannerCommPort )
		{
			// Usually, we should get a WM_QUIT before the COMM port closes, 
			// but if not, go ahead and exit...
			ROBOT_LOG( TRUE, "LaserScannerCommWriteThreadFunc: Port INVALID HANDLE \n")
			return 0;	
		}

		if( (msg.message < WM_ROBOT_MESSAGE_BASE) || ((msg.message > (WM_ROBOT_MESSAGE_BASE+HW_MAX_MESSAGE) ) ) )
		{
			// Ignore any messages not intended for user (where do these come from?)
			ROBOT_LOG( TRUE, "LaserScannerCommWriteThreadFunc: message out of Range, WM_ROBOT_MESSAGE_BASE = 0x%08X, Msg = 0x%08lX, wParam = 0x%08lX, lParam = 0x%08lX\n", 
				WM_ROBOT_MESSAGE_BASE, msg.message, msg.wParam, msg.lParam )
			continue;
		}

		{
			//TAL_Event("Handle Laser Command");
			__itt_task_begin(pDomainLaserThread, __itt_null, __itt_null, pshLaserHandleCommand);
			LaserScannerCommand.HandleCommand( Request, (msg.wParam), (msg.lParam) );
			__itt_task_end(pDomainLaserThread);
		}

		__itt_task_end(pDomainLaserThread);  // pshLaserWriteThreadLoop
		Sleep(0); // force thread switch after each command
	}
	ROBOT_LOG( TRUE, "Laser Scanner Control Thread: Received WM_ROBOT_THREAD_EXIT_CMD\n")
	CloseCommPort( "Laser" , g_hLaserScannerCommPort );

	return 0;
}

////////////////////////////////////////////////////////////////////////////////////////
__itt_string_handle* pshLaserReadThreadLoop = __itt_string_handle_create("Laser Read Loop");
__itt_string_handle* pshLaserParseBuffer = __itt_string_handle_create("Parse Buffer");
DWORD WINAPI LaserScannerCommReadThreadFunc(LPVOID lpParameter) // g_hLaserScannerCommPort
{
	///TAL_SetThreadName("Laser Read Thread");
	IGNORE_UNUSED_PARAM (lpParameter);
	__itt_thread_set_name( "Laser Read Thread" );

//	HANDLE hCommPort = (HANDLE)lpParameter;	// Use Global instead
//	DWORD	fdwCommMask;
	DWORD	dwSIOBytesReceived = 0;
	//char	LaserSIOBuf[LASER_SCANNER_READ_BUF_SIZE+1];

	char *LaserSIOBuf = new char[LASER_SCANNER_READ_BUF_SIZE+1];
	memset( LaserSIOBuf,0,LASER_SCANNER_READ_BUF_SIZE );

	// Create the Laser Parser object
	CLaserScannerParser *pLaserScannerParser = new CLaserScannerParser;
	pLaserScannerParser->Init();

	DWORD dwBytesReceived;

	if( SIMULATED_SIO_HANDLE != g_hPicCommPort )
	{
		SetCommMask( g_hLaserScannerCommPort, EV_RXCHAR | EV_BREAK );
	}
	Sleep(1000);	// Alow other initialization to occur first

	while(g_hLaserScannerCommPort != INVALID_HANDLE_VALUE)
	{
		///TAL_Event("Laser Read Loop Top");
		///TAL_SCOPED_TASK_NAMED("Laser Read Loop");
		__itt_task_begin(pDomainLaserThread, __itt_null, __itt_null, pshLaserReadThreadLoop);

		dwBytesReceived = 0;
		LaserSIOBuf[0] = '\0';
		//memset( LaserSIOBuf,0,LASER_SCANNER_READ_BUF_SIZE );	// DEBUG
		//memset( SioBuffer, '\0', PIC_SIO_BUF_LEN );	// clean out garbage chars


		BOOL bSucess = 0;
		{
			///TAL_SCOPED_TASK_NAMED("Read LsrSIO");

			// Read ANSI characters
			if( SIMULATED_SIO_HANDLE == g_hLaserScannerCommPort )
			{
				Sleep(SIMULATED_SIO_SLEEP_TIME);
				continue;
			}

			bSucess = ReadFile(g_hLaserScannerCommPort,	// where to read from (the open comm port)
					 LaserSIOBuf,						// where to put the character
					 (LASER_SCANNER_READ_BUF_SIZE-1),	// number of bytes to read
					 &dwSIOBytesReceived,				// how many bytes were actually read
					 NULL);								// no overlapped I/O
		}
		if( !bSucess ) 
		{

			ROBOT_DISPLAY( TRUE, "ERROR Reading Laser Comm port\n" )
			PurgeComm(g_hLaserScannerCommPort, PURGE_RXABORT|PURGE_RXCLEAR);
			//PurgeComm(hCommPort, PURGE_RXABORT|PURGE_TXABORT|PURGE_RXCLEAR|PURGE_TXCLEAR);
			DWORD dwCommError = 0;
			ClearCommError(g_hLaserScannerCommPort, &dwCommError, NULL);
			ReportCommError("Laser SIO Read Error", dwCommError );

			return 0;		// terminate thread on error
		}

		if( 0 != dwSIOBytesReceived )
		{
			//ROBOT_LOG( TRUE,"LASER Read Thread: {%s}", LaserSIOBuf )
			//ROBOT_LOG( TRUE,"LASER Read Thread: %d bytes received\n", dwSIOBytesReceived )

			// Got some Laser data.  Parse it.
			__itt_task_begin(pDomainLaserThread, __itt_null, __itt_null, pshLaserParseBuffer);
			if( pLaserScannerParser->ParseBuffer( LaserSIOBuf, dwSIOBytesReceived ) )
			{
				// Indicate that scan is complete
				gWaitingForLaserSingleLineScan = FALSE;

				// Now throw message in the queue, to indicate that the Laser data has been updated
				//TODO!	PostThreadMessage( g_dwControlThreadId, (WM_ROBOT_GPS_DATA_READY), 0, 0 );
				// Post message to the GUI display (local or remote)
				SendResponse( WM_ROBOT_DISPLAY_SINGLE_ITEM,	// command
					ROBOT_RESPONSE_LASER_SCANNER_DATA,			// Param1 = Bulk data command
					0 );										// Param2 = not used
			}
			__itt_task_end(pDomainLaserThread);  // pshLaserParseBuffer
		}

		__itt_task_end(pDomainLaserThread);  // pshLaserReadThreadLoop
		if( ! g_bLaserContinuousScanEnabled )
		{
			Sleep(100); // just check every now and then, since auto laser not requested
		}
		else
		{
			// NOTE: Timing verified.  Don't check faster than this, as the laser needs time to get the data after a command is sent
			Sleep(20); // don't spin in a tight loop!
		}
	}	// while port open


	ROBOT_LOG( TRUE,"LASER SCANNER COMM Read Thread exiting.\n")
	SAFE_DELETE( pLaserScannerParser );
	SAFE_DELETE( LaserSIOBuf );
	return 0;
}




#endif // ROBOT_SERVER - nothing in this file used for client 

