/************************************************************************
*                                                                       *
*   NuiImpl.cpp -- Implementation of CKinectNui methods dealing *
*                  with NUI processing                                  *
*                                                                       *
* Copyright (c) Microsoft Corp. All rights reserved.                    *
*                                                                       *
* This code is licensed under the terms of the                          *
* Microsoft Kinect for Windows SDK (Beta) from Microsoft Research       *
* License Agreement: http://research.microsoft.com/KinectSDK-ToU        *
*                                                                       *
************************************************************************/

#include "stdafx.h"
#include "Globals.h"
#include "Kinect.h"
#include "HardwareConfig.h"
#include "resource.h"
#include <mmsystem.h>

#if ( KINECT_SDK_TYPE_MICROSOFT == 1 )
// Everything in this file is for Microsoft SDK use only!

static const COLORREF g_JointColorTable[NUI_SKELETON_POSITION_COUNT] = 
{
    RGB(169, 176, 155), // NUI_SKELETON_POSITION_HIP_CENTER
    RGB(169, 176, 155), // NUI_SKELETON_POSITION_SPINE
    RGB(168, 230, 29), // NUI_SKELETON_POSITION_SHOULDER_CENTER
    RGB(200, 0,   0), // NUI_SKELETON_POSITION_HEAD
    RGB(79,  84,  33), // NUI_SKELETON_POSITION_SHOULDER_LEFT
    RGB(84,  33,  42), // NUI_SKELETON_POSITION_ELBOW_LEFT
    RGB(255, 126, 0), // NUI_SKELETON_POSITION_WRIST_LEFT
    RGB(215,  86, 0), // NUI_SKELETON_POSITION_HAND_LEFT
    RGB(33,  79,  84), // NUI_SKELETON_POSITION_SHOULDER_RIGHT
    RGB(33,  33,  84), // NUI_SKELETON_POSITION_ELBOW_RIGHT
    RGB(77,  109, 243), // NUI_SKELETON_POSITION_WRIST_RIGHT
    RGB(37,   69, 243), // NUI_SKELETON_POSITION_HAND_RIGHT
    RGB(77,  109, 243), // NUI_SKELETON_POSITION_HIP_LEFT
    RGB(69,  33,  84), // NUI_SKELETON_POSITION_KNEE_LEFT
    RGB(229, 170, 122), // NUI_SKELETON_POSITION_ANKLE_LEFT
    RGB(255, 126, 0), // NUI_SKELETON_POSITION_FOOT_LEFT
    RGB(181, 165, 213), // NUI_SKELETON_POSITION_HIP_RIGHT
    RGB(71, 222,  76), // NUI_SKELETON_POSITION_KNEE_RIGHT
    RGB(245, 228, 156), // NUI_SKELETON_POSITION_ANKLE_RIGHT
    RGB(77,  109, 243) // NUI_SKELETON_POSITION_FOOT_RIGHT
};




void CKinectNui::Nui_Zero()
{
    m_hNextDepthFrameEvent = NULL;
    m_hNextVideoFrameEvent = NULL;
    m_hNextSkeletonEvent = NULL;
    m_pDepthStreamHandle = NULL;
    m_pVideoStreamHandle = NULL;
    //m_hThNuiProcess = NULL;
    m_hEvNuiProcessStop = NULL;
    ZeroMemory(m_Pen,sizeof(m_Pen));
    m_SkeletonDC = NULL;
    m_SkeletonBMP = NULL;
    m_SkeletonOldObj = NULL;
    m_PensTotal = 6;
    ZeroMemory(m_Points,sizeof(m_Points));
    m_LastSkeletonFoundTime = -1;
    m_bScreenBlanked = false;
    m_FramesTotal = 0;
    m_LastFPStime = -1;
    m_LastFramesTotal = 0;
}



HRESULT CKinectNui::Nui_Init()
{
    HRESULT                hr;
    RECT                rc;

    m_hNextDepthFrameEvent = CreateEvent( NULL, TRUE, FALSE, NULL );
    m_hNextVideoFrameEvent = CreateEvent( NULL, TRUE, FALSE, NULL );
    m_hNextSkeletonEvent = CreateEvent( NULL, TRUE, FALSE, NULL );

    GetWindowRect(GetDlgItem( m_hWnd, IDC_SKELETAL_VIEW ), &rc );
    int width = rc.right - rc.left;
    int height = rc.bottom - rc.top;
    HDC hdc = GetDC(GetDlgItem( m_hWnd, IDC_SKELETAL_VIEW));
    m_SkeletonBMP = CreateCompatibleBitmap( hdc, width, height );
    m_SkeletonDC = CreateCompatibleDC( hdc );
    ::ReleaseDC(GetDlgItem(m_hWnd,IDC_SKELETAL_VIEW), hdc );
    m_SkeletonOldObj = SelectObject( m_SkeletonDC, m_SkeletonBMP );

	HWND hwndDepth = GetDlgItem( m_hWnd, IDC_DEPTH_VIEW );

    hr = m_DrawDepth.CreateDevice( hwndDepth );
    //hr = m_DrawDepth.CreateDevice( GetDlgItem( m_hWnd, IDC_DEPTH_VIEW ) );
    if( FAILED( hr ) )
    {
		ROBOT_LOG( TRUE, "*************** KINECT ERROR: Could not create Direct3D DEPTH surface. Ensure the DirectX runtime is installed.  ***************" )
        return hr;
    }

    hr = m_DrawDepth.SetVideoType( 320, 240, 320 * 4 );
    if( FAILED( hr ) )
    {
		ROBOT_LOG( TRUE, "KINECT ERROR: Could not set Direct3D video type." )
        //MessageBoxResource( m_hWnd,IDS_ERROR_D3DVIDEOTYPE,MB_OK | MB_ICONHAND);
        return hr;
    }

    hr = m_DrawVideo.CreateDevice( GetDlgItem( m_hWnd, IDC_VIDEO_VIEW ) );
    if( FAILED( hr ) )
    {
		ROBOT_LOG( TRUE, "*************** KINECT ERROR: Could not create Direct3D VIDEO surface. Ensure the DirectX runtime is installed. ***************" )
        //MessageBoxResource( m_hWnd,IDS_ERROR_D3DCREATE,MB_OK | MB_ICONHAND);
        return hr;
    }

    hr = m_DrawVideo.SetVideoType( 640, 480, 640 * 4 );
    if( FAILED( hr ) )
    {
		ROBOT_LOG( TRUE, "*************** KINECT ERROR: Could not set Direct3D video type.  ***************" )
        //MessageBoxResource( m_hWnd,IDS_ERROR_D3DVIDEOTYPE,MB_OK | MB_ICONHAND);
        return hr;
    }
    
    hr = NuiInitialize( 
        NUI_INITIALIZE_FLAG_USES_DEPTH_AND_PLAYER_INDEX | NUI_INITIALIZE_FLAG_USES_SKELETON | NUI_INITIALIZE_FLAG_USES_COLOR);
    if( FAILED( hr ) )
    {
		ROBOT_LOG( TRUE, "*************** KINECT ERROR: NuiInitialize failed ***************" )
        //MessageBoxResource(m_hWnd,IDS_ERROR_NUIINIT,MB_OK | MB_ICONHAND);
        return hr;
    }

    hr = NuiSkeletonTrackingEnable( m_hNextSkeletonEvent, 0 );
    if( FAILED( hr ) )
    {
		ROBOT_LOG( TRUE, "***************  KINECT ERROR: Could not enable skeleton tracking ***************" )
       //MessageBoxResource(m_hWnd,IDS_ERROR_SKELETONTRACKING,MB_OK | MB_ICONHAND);
        return hr;
    }

    hr = NuiImageStreamOpen(
        NUI_IMAGE_TYPE_COLOR,
        NUI_IMAGE_RESOLUTION_640x480,
        0,
        2,
        m_hNextVideoFrameEvent,
        &m_pVideoStreamHandle );
    if( FAILED( hr ) )
    {
		ROBOT_LOG( TRUE, "*************** KINECT ERROR: Could not open image stream video ***************" )
        //MessageBoxResource(m_hWnd,IDS_ERROR_VIDEOSTREAM,MB_OK | MB_ICONHAND);
        return hr;
    }

    hr = NuiImageStreamOpen(
        NUI_IMAGE_TYPE_DEPTH_AND_PLAYER_INDEX,
        NUI_IMAGE_RESOLUTION_320x240,
        0,
        2,
        m_hNextDepthFrameEvent,
        &m_pDepthStreamHandle );
    if( FAILED( hr ) )
    {
		ROBOT_LOG( TRUE, "*************** KINECT ERROR: Could not open image stream depth  ***************" )
        //MessageBoxResource(m_hWnd,IDS_ERROR_DEPTHSTREAM,MB_OK | MB_ICONHAND);
        return hr;
    }

    // Start the Nui processing thread
    m_hEvNuiProcessStop = CreateEvent( NULL, FALSE, FALSE, NULL );
    g_hKinectNuiThread = CreateThread( NULL, 0, Nui_ProcessThread, this, 0, NULL );

    return hr;
}



void CKinectNui::Nui_UnInit( )
{
    ::SelectObject( m_SkeletonDC, m_SkeletonOldObj );
    DeleteDC( m_SkeletonDC );
    DeleteObject( m_SkeletonBMP );

    if( m_Pen[0] != NULL )
    {
        DeleteObject(m_Pen[0]);
        DeleteObject(m_Pen[1]);
        DeleteObject(m_Pen[2]);
        DeleteObject(m_Pen[3]);
        DeleteObject(m_Pen[4]);
        DeleteObject(m_Pen[5]);
        ZeroMemory(m_Pen,sizeof(m_Pen));
    }

    // Stop the Nui processing thread
    if(m_hEvNuiProcessStop != NULL)
    {
        // Signal the thread
        SetEvent(m_hEvNuiProcessStop);

        // Wait for thread to stop
        if( g_hKinectNuiThread != INVALID_HANDLE_VALUE )
        {
            WaitForSingleObject( g_hKinectNuiThread, INFINITE ); //1000?
            CloseHandle( g_hKinectNuiThread );
        }
        CloseHandle( m_hEvNuiProcessStop );
    }

    NuiShutdown( );
    if( m_hNextSkeletonEvent && ( m_hNextSkeletonEvent != INVALID_HANDLE_VALUE ) )
    {
        CloseHandle( m_hNextSkeletonEvent );
        m_hNextSkeletonEvent = NULL;
    }
    if( m_hNextDepthFrameEvent && ( m_hNextDepthFrameEvent != INVALID_HANDLE_VALUE ) )
    {
        CloseHandle( m_hNextDepthFrameEvent );
        m_hNextDepthFrameEvent = NULL;
    }
    if( m_hNextVideoFrameEvent && ( m_hNextVideoFrameEvent != INVALID_HANDLE_VALUE ) )
    {
        CloseHandle( m_hNextVideoFrameEvent );
        m_hNextVideoFrameEvent = NULL;
    }
    m_DrawDepth.DestroyDevice( );
    m_DrawVideo.DestroyDevice( );
}



DWORD WINAPI CKinectNui::Nui_ProcessThread(LPVOID pParam)
{
    CKinectNui *pthis=(CKinectNui *) pParam;
    HANDLE                hEvents[4];
    int                    nEventIdx,t,dt;

    // Configure events to be listened on
    hEvents[0]=pthis->m_hEvNuiProcessStop;
    hEvents[1]=pthis->m_hNextDepthFrameEvent;
    hEvents[2]=pthis->m_hNextVideoFrameEvent;
    hEvents[3]=pthis->m_hNextSkeletonEvent;

    pthis->Nui_BlankSkeletonScreen( GetDlgItem( pthis->m_hWnd, IDC_SKELETAL_VIEW ) );

    // Main thread loop
    while(1)
    {
        // Wait for an event to be signalled
        nEventIdx=WaitForMultipleObjects(sizeof(hEvents)/sizeof(hEvents[0]),hEvents,FALSE,100);

        // If the stop event, stop looping and exit
        if(nEventIdx==0)
            break;            

        // Perform FPS processing
        t = timeGetTime( );
        if( pthis->m_LastFPStime == -1 )
        {
            pthis->m_LastFPStime = t;
            pthis->m_LastFramesTotal = pthis->m_FramesTotal;
        }
        dt = t - pthis->m_LastFPStime;
        if( dt > 1000 )
        {
            pthis->m_LastFPStime = t;
            int FrameDelta = pthis->m_FramesTotal - pthis->m_LastFramesTotal;
            pthis->m_LastFramesTotal = pthis->m_FramesTotal;
 // TODO - PRINT OUT FPS ON THE GUI:   SetDlgItemInt( pthis->m_hWnd, IDC_FPS, FrameDelta,FALSE );
        }

        // Perform skeletal panel blanking
        if( pthis->m_LastSkeletonFoundTime == -1 )
            pthis->m_LastSkeletonFoundTime = t;
        dt = t - pthis->m_LastSkeletonFoundTime;
        if( dt > 250 )
        {
            if( !pthis->m_bScreenBlanked )
            {
                pthis->Nui_BlankSkeletonScreen( GetDlgItem( pthis->m_hWnd, IDC_SKELETAL_VIEW ) );
                pthis->m_bScreenBlanked = true;
            }
        }

        // Process signal events
        switch(nEventIdx)
        {
            case 1:
                pthis->Nui_GotDepthAlert();
                pthis->m_FramesTotal++;
                break;

            case 2:
                pthis->Nui_GotVideoAlert();
                break;

            case 3:
                pthis->Nui_GotSkeletonAlert( );
                break;
        }
    }

    return (0);
}

void CKinectNui::Nui_GotVideoAlert( )
{
    const NUI_IMAGE_FRAME * pImageFrame = NULL;

    HRESULT hr = NuiImageStreamGetNextFrame(
        m_pVideoStreamHandle,
        0,
        &pImageFrame );
    if( FAILED( hr ) )
    {
        return;
    }

    NuiImageBuffer * pTexture = pImageFrame->pFrameTexture;
    KINECT_LOCKED_RECT LockedRect;
    pTexture->LockRect( 0, &LockedRect, NULL, 0 );
    if( LockedRect.Pitch != 0 )
    {
        BYTE * pBuffer = (BYTE*) LockedRect.pBits;

        m_DrawVideo.DrawFrame( (BYTE*) pBuffer );

    }
    else
    {
        ROBOT_LOG( TRUE, "Buffer length of received texture is bogus\n" )
    }

    NuiImageStreamReleaseFrame( m_pVideoStreamHandle, pImageFrame );
}


void CKinectNui::Nui_GotDepthAlert( )
{
	///////////////////////////////////////////////////////////////////////////////////////////
	// Handle Depth updates here!
	///////////////////////////////////////////////////////////////////////////////////////////

    const NUI_IMAGE_FRAME * pImageFrame = NULL;

    HRESULT hr = NuiImageStreamGetNextFrame(
        m_pDepthStreamHandle,
        0,
        &pImageFrame );

    if( FAILED( hr ) )
    {
        return;
    }

    NuiImageBuffer * pTexture = pImageFrame->pFrameTexture;
    KINECT_LOCKED_RECT LockedRect;
    pTexture->LockRect( 0, &LockedRect, NULL, 0 );
    if( LockedRect.Pitch != 0 )
    {
        BYTE * pBuffer = (BYTE*) LockedRect.pBits;


		// Convert depth data to colors and map to Depth frame
		// Starts at top of frame (y=0) and scans to bottom of frame (y=frameheight)
		// So, top of frame is farthest from the robot!
		// NOTE: Bottom 3 pixels indicates Player (Human detected) number!

		if( NULL == g_KinectPointCloud )
		{
			ROBOT_ASSERT(0); // Not initilized!
			return;
		}

		KINECT_SURFACE_DESC FrameDesc;
		pTexture->GetLevelDesc(0, &FrameDesc);

		/// double ScaleFactor = 5.0;
		/// TODO - Determine max range for scale (from KinectModule.cpp)

        // draw the bits to the bitmap
        RGBQUAD * rgbrun = m_rgbWk;
        USHORT * pBufferRun = (USHORT*) pBuffer;
        for( int y = 0 ; y < 240 ; y++ )
        {
            for( int x = 0 ; x < 320 ; x++ )
            {
				if( (120 == y) && (160 == x) )
				{ // center pixel
					USHORT s = *pBufferRun;
					USHORT DistMM = (s & 0xfff8) >> 3;
					ROBOT_LOG( TRUE, "DEBUG DEPTH BUFFER: Center Pixel = %d mm\n", DistMM )
					
					Nui_UpdatePointCloud( x, y, DistMM, FrameDesc ); // Update the 3D point cloud
				}
                // MODE 1: 
				RGBQUAD quad = Nui_ShortToQuad_Depth( *pBufferRun );
                // MODE 2:
				//RGBQUAD quad = Nui_ShortToColorQuad_Depth( *pBufferRun );
                pBufferRun++;
                *rgbrun = quad;
                rgbrun++;
            }
        }

        m_DrawDepth.DrawFrame( (BYTE*) m_rgbWk );
    }
    else
    {
       ROBOT_LOG( TRUE, "Buffer length of received texture is bogus\n" )
    }

    NuiImageStreamReleaseFrame( m_pDepthStreamHandle, pImageFrame );
}



void CKinectNui::Nui_BlankSkeletonScreen(HWND hWnd)
{
    HDC hdc = GetDC( hWnd );
    RECT rct;
    GetClientRect(hWnd, &rct);
    int width = rct.right;
    int height = rct.bottom;
    PatBlt( hdc, 0, 0, width, height, BLACKNESS );
    ReleaseDC( hWnd, hdc );
}

void CKinectNui::Nui_DrawSkeletonSegment( NUI_SKELETON_DATA * pSkel, int numJoints, ... )
{
    va_list vl;
    va_start(vl,numJoints);
    POINT segmentPositions[NUI_SKELETON_POSITION_COUNT];

    for (int iJoint = 0; iJoint < numJoints; iJoint++)
    {
        NUI_SKELETON_POSITION_INDEX jointIndex = va_arg(vl,NUI_SKELETON_POSITION_INDEX);
        segmentPositions[iJoint].x = m_Points[jointIndex].x;
        segmentPositions[iJoint].y = m_Points[jointIndex].y;
    }

    Polyline(m_SkeletonDC, segmentPositions, numJoints);

    va_end(vl);
}

void CKinectNui::Nui_DrawSkeleton( bool bBlank, NUI_SKELETON_DATA * pSkel, HWND hWnd, int WhichSkeletonColor )
{
    HGDIOBJ hOldObj = SelectObject(m_SkeletonDC,m_Pen[WhichSkeletonColor % m_PensTotal]);
    
    RECT rct;
    GetClientRect(hWnd, &rct);
    int width = rct.right;
    int height = rct.bottom;

    if( m_Pen[0] == NULL )
    {
        m_Pen[0] = CreatePen( PS_SOLID, width / 80, RGB(255, 0, 0) );
        m_Pen[1] = CreatePen( PS_SOLID, width / 80, RGB( 0, 255, 0 ) );
        m_Pen[2] = CreatePen( PS_SOLID, width / 80, RGB( 64, 255, 255 ) );
        m_Pen[3] = CreatePen( PS_SOLID, width / 80, RGB(255, 255, 64 ) );
        m_Pen[4] = CreatePen( PS_SOLID, width / 80, RGB( 255, 64, 255 ) );
        m_Pen[5] = CreatePen( PS_SOLID, width / 80, RGB( 128, 128, 255 ) );
    }

    if( bBlank )
    {
        PatBlt( m_SkeletonDC, 0, 0, width, height, BLACKNESS );
    }

    int scaleX = width; //scaling up to image coordinates
    int scaleY = height;
    float fx=0,fy=0;
    int i;
    for (i = 0; i < NUI_SKELETON_POSITION_COUNT; i++)
    {
        NuiTransformSkeletonToDepthImageF( pSkel->SkeletonPositions[i], &fx, &fy );
        m_Points[i].x = (int) ( fx * scaleX + 0.5f );
        m_Points[i].y = (int) ( fy * scaleY + 0.5f );
    }

    SelectObject(m_SkeletonDC,m_Pen[WhichSkeletonColor%m_PensTotal]);
    
    Nui_DrawSkeletonSegment(pSkel,4,NUI_SKELETON_POSITION_HIP_CENTER, NUI_SKELETON_POSITION_SPINE, NUI_SKELETON_POSITION_SHOULDER_CENTER, NUI_SKELETON_POSITION_HEAD);
    Nui_DrawSkeletonSegment(pSkel,5,NUI_SKELETON_POSITION_SHOULDER_CENTER, NUI_SKELETON_POSITION_SHOULDER_LEFT, NUI_SKELETON_POSITION_ELBOW_LEFT, NUI_SKELETON_POSITION_WRIST_LEFT, NUI_SKELETON_POSITION_HAND_LEFT);
    Nui_DrawSkeletonSegment(pSkel,5,NUI_SKELETON_POSITION_SHOULDER_CENTER, NUI_SKELETON_POSITION_SHOULDER_RIGHT, NUI_SKELETON_POSITION_ELBOW_RIGHT, NUI_SKELETON_POSITION_WRIST_RIGHT, NUI_SKELETON_POSITION_HAND_RIGHT);
    Nui_DrawSkeletonSegment(pSkel,5,NUI_SKELETON_POSITION_HIP_CENTER, NUI_SKELETON_POSITION_HIP_LEFT, NUI_SKELETON_POSITION_KNEE_LEFT, NUI_SKELETON_POSITION_ANKLE_LEFT, NUI_SKELETON_POSITION_FOOT_LEFT);
    Nui_DrawSkeletonSegment(pSkel,5,NUI_SKELETON_POSITION_HIP_CENTER, NUI_SKELETON_POSITION_HIP_RIGHT, NUI_SKELETON_POSITION_KNEE_RIGHT, NUI_SKELETON_POSITION_ANKLE_RIGHT, NUI_SKELETON_POSITION_FOOT_RIGHT);
    
    // Draw the joints in a different color
    for (i = 0; i < NUI_SKELETON_POSITION_COUNT ; i++)
    {
        HPEN hJointPen;
        
        hJointPen=CreatePen(PS_SOLID,9, g_JointColorTable[i]);
        hOldObj=SelectObject(m_SkeletonDC,hJointPen);

        MoveToEx( m_SkeletonDC, m_Points[i].x, m_Points[i].y, NULL );
        LineTo( m_SkeletonDC, m_Points[i].x, m_Points[i].y );

        SelectObject( m_SkeletonDC, hOldObj );
        DeleteObject(hJointPen);
    }

    return;

}




void CKinectNui::Nui_DoDoubleBuffer(HWND hWnd,HDC hDC)
{
    RECT rct;
    GetClientRect(hWnd, &rct);
    int width = rct.right;
    int height = rct.bottom;

    HDC hdc = GetDC( hWnd );

    BitBlt( hdc, 0, 0, width, height, hDC, 0, 0, SRCCOPY );

    ReleaseDC( hWnd, hdc );

}

void CKinectNui::Nui_GotSkeletonAlert( )
{
    NUI_SKELETON_FRAME SkeletonFrame;

    HRESULT hr = NuiSkeletonGetNextFrame( 0, &SkeletonFrame );

    bool bFoundSkeleton = true;
    for( int i = 0 ; i < NUI_SKELETON_COUNT ; i++ )
    {
        if( SkeletonFrame.SkeletonData[i].eTrackingState == NUI_SKELETON_TRACKED )
        {
            bFoundSkeleton = false;
        }
    }

    // no skeletons!
    //
    if( bFoundSkeleton )
    {
        return;
    }

    // smooth out the skeleton data
    NuiTransformSmooth(&SkeletonFrame,NULL);

    // we found a skeleton, re-start the timer
    m_bScreenBlanked = false;
    m_LastSkeletonFoundTime = -1;

    // draw each skeleton color according to the slot within they are found.
    //
    bool bBlank = true;
    for( int i = 0 ; i < NUI_SKELETON_COUNT ; i++ )
    {
        if( SkeletonFrame.SkeletonData[i].eTrackingState == NUI_SKELETON_TRACKED )
        {
            Nui_DrawSkeleton( bBlank, &SkeletonFrame.SkeletonData[i], GetDlgItem( m_hWnd, IDC_SKELETAL_VIEW ), i );
            bBlank = false;
        }
    }

    Nui_DoDoubleBuffer(GetDlgItem(m_hWnd,IDC_SKELETAL_VIEW), m_SkeletonDC);
}



RGBQUAD CKinectNui::Nui_ShortToQuad_Depth( USHORT s )
{
    USHORT RealDepth = (s & 0xfff8) >> 3;
    USHORT Player = s & 7; // 3 bottom bits indicates the player number!

    // transform 13-bit depth information into an 8-bit intensity appropriate
    // for display (we disregard information in most significant bit)
    BYTE l = 255 - (BYTE)(256*RealDepth/0x0fff);

    RGBQUAD q;
    q.rgbRed = q.rgbBlue = q.rgbGreen = 0;

    switch( Player )
    {
    case 0:
        q.rgbRed = l / 2;
        q.rgbBlue = l / 2;
        q.rgbGreen = l / 2;
        break;
    case 1:
        q.rgbRed = l;
        break;
    case 2:
        q.rgbGreen = l;
        break;
    case 3:
        q.rgbRed = l / 4;
        q.rgbGreen = l;
        q.rgbBlue = l;
        break;
    case 4:
        q.rgbRed = l;
        q.rgbGreen = l;
        q.rgbBlue = l / 4;
        break;
    case 5:
        q.rgbRed = l;
        q.rgbGreen = l / 4;
        q.rgbBlue = l;
        break;
    case 6:
        q.rgbRed = l / 2;
        q.rgbGreen = l / 2;
        q.rgbBlue = l;
        break;
    case 7:
        q.rgbRed = 255 - ( l / 2 );
        q.rgbGreen = 255 - ( l / 2 );
        q.rgbBlue = 255 - ( l / 2 );
    }

    return q;
}

RGBQUAD CKinectNui::Nui_ShortToColorQuad_Depth( USHORT s )
{
	// DAVES: Uses my unique method for color depth visualization
    RGBQUAD Pixel;
    USHORT RealDepth = (s & 0xfff8) >> 3;
    USHORT Player = s & 7; // 3 bottom bits indicates the player number!
	UINT DepthValueMM = RealDepth;	// Depth value in millimeters

    // transform 13-bit depth information into an 8-bit intensity appropriate
    // for display (we disregard information in most significant bit)
    BYTE l = 255 - (BYTE)(256*RealDepth/0x0fff);

	double ScaleFactor = 5;
    Pixel.rgbRed = Pixel.rgbBlue = Pixel.rgbGreen = 0;

	if( 0 != Player )
	{
		// Use solid colors for human players!
	   switch( Player )
		{
			case 1:
				Pixel.rgbRed = l;
				break;
			case 2:
				Pixel.rgbGreen = l;
				break;
			case 3:
				Pixel.rgbRed = l / 4;
				Pixel.rgbGreen = l;
				Pixel.rgbBlue = l;
				break;
			case 4:
				Pixel.rgbRed = l;
				Pixel.rgbGreen = l;
				Pixel.rgbBlue = l / 4;
				break;
			case 5:
				Pixel.rgbRed = l;
				Pixel.rgbGreen = l / 4;
				Pixel.rgbBlue = l;
				break;
			case 6:
				Pixel.rgbRed = l / 2;
				Pixel.rgbGreen = l / 2;
				Pixel.rgbBlue = l;
				break;
			case 7:
				Pixel.rgbRed = 255 - ( l / 2 );
				Pixel.rgbGreen = 255 - ( l / 2 );
				Pixel.rgbBlue = 255 - ( l / 2 );
		}
	}
	else if( 0 != DepthValueMM  )	
	{
		UINT red =0; UINT blue =0; UINT green =0; 
		signed int ScaledValue = (signed int) (((double)DepthValueMM / ScaleFactor) -90.0); // scale, and subtract out min range (47 ScaledValue)
		if( ScaledValue < 0) ScaledValue = 0;
		UINT modu = ScaledValue % 128;

		if( ScaledValue < 128.0 )
		{
			green = 256-modu;	// Ramp down
		}
		else if( ScaledValue < 256 )
		{
			green = 128-modu;	// Ramp down
			blue = modu;		// Ramp up
		}
		else if( ScaledValue < 384 )
		{
			blue = modu+128;	// Ramp up
		}
		else if( ScaledValue < 512 )
		{
			red = modu;			// Ramp up
			blue = 256-modu;	// Ramp down
		}
		else if( ScaledValue < 640 )
		{
			red = modu+128;		// Ramp up
			blue = 128-modu;	// Ramp down
		}
		else if( ScaledValue < 768 )
		{
			red = 256-modu;		// Ramp down
		}
		else if( ScaledValue < 896 )
		{
			red = 128-modu;		// Ramp down
		}
		else if( ScaledValue < 964 )
		{
			red = 128-modu;		// Ramp down
			green = 128-modu;	// Ramp down
			blue = 128-modu;	// Ramp down
		}
		else 
		{
			red = 60;	
			green = 60;	
			blue = 60;
		}

		Pixel.rgbRed = red;
		Pixel.rgbGreen = green;
		Pixel.rgbBlue = blue;
	}

    return Pixel;
}


void CKinectNui::Nui_UpdatePointCloud( UINT x, UINT y, int DepthValueMM, KINECT_SURFACE_DESC FrameDesc )
{
	// Update the 3D point cloud
	// Calculate X,Y,Z based upon tilt of Kinect and FOV of the Kinect Camera
	double FrameWidth = (double)FrameDesc.Width;
	double FrameHeight = (double)FrameDesc.Height;

	// Calculate degrees per pixel for the image
	double DegreeIncrementX = KINECT_DEPTH_FOV_X / FrameWidth;
	double DegreeIncrementY = KINECT_DEPTH_FOV_Y / FrameHeight;

	// Get Kinect Tilt angle, and position offset caused by tilt
	double KinectTiltTenthDegrees =	g_BulkServoStatus[DYNA_KINECT_SCANNER_SERVO_ID].PositionTenthDegrees - 10.0;	// Add a small fudge factor
	double HeightAboveFloor = KINECT_HEIGHT_ABOVE_GROUND_TENTH_INCHES;
	double DistanceFromFront = KINECT_DISTANCE_FROM_FRONT_TENTH_INCHES; 

	// x, y = pixel location
	double X=0,Y=0,Z=0,TempX=0,TempY=0,TempZ=0,m=0,n=0,q=0,u=0, RangeTenthInches = 0;

	/*
	if( m_DepthMousePointSelected )
	{
		if( (x == MouseX) && (y == MouseY) )
		{
			ROBOT_LOG( TRUE,  "MOUSE DOWN  ")
		}
	}
	*/

	if( (DepthValueMM >= KINECT_MM_MAX) || (DepthValueMM <= 0) )
	{
		//ROBOT_LOG( TRUE,  "Kinect Read Zero: EnterCriticalSection g_csKinectPointCloudLock\n" )
		__itt_task_begin(pDomainKinectThread, __itt_null, __itt_null, psh_csKinectPointCloudLock);
		EnterCriticalSection(&g_csKinectPointCloudLock);
			g_KinectPointCloud->Point3dArray[y][x].X = KINECT_RANGE_TENTH_INCHES_MAX;
			g_KinectPointCloud->Point3dArray[y][x].Y = KINECT_RANGE_TENTH_INCHES_MAX;
			g_KinectPointCloud->Point3dArray[y][x].Z = KINECT_RANGE_TENTH_INCHES_MAX;
		LeaveCriticalSection(&g_csKinectPointCloudLock);
		__itt_task_end(pDomainKinectThread);
		//ROBOT_LOG( TRUE,  "Kinect Read Zero: LeaveCriticalSection g_csKinectPointCloudLock\n" )
	}
	else
	{
		RangeTenthInches = ((double)DepthValueMM / 2.540);	

		double StepX = (double)x - ((double)FrameWidth/2.0);	// goes from (-320 to + 320)
		double StepAngleX = DegreeIncrementX / 1.5 * StepX; // goes from (-46 to + 46 degrees)

		double StepY = ( ((double)FrameHeight/2.0) - (double)y );	// goes from (+240 to -240)
		//double StepAngleY =  DegreeIncrementY * StepY; // goes from (-38.6 to + 38.6 degrees)
		double StepAngleY =  (KinectTiltTenthDegrees/10.0) + (DegreeIncrementY / 1.6 * StepY); // goes from (-38.6 to + 38.6 degrees) + Tilt angle
		double StepAngleYZ =  (KinectTiltTenthDegrees/10.0) + (DegreeIncrementY / 1.8 * StepY); // goes from (-38.6 to + 38.6 degrees) + Tilt angle
		//double StepAngleY =  (KinectTiltTenthDegrees/10.0); // goes from (-38.6 to + 38.6 degrees) + Tilt angle

		// Get x,y,z from Kinect's perspective. zero = straight forward facing.  Down = negative
		// Convert from Spherical Coordinates to X,Y,Z coordinates for each point
		TempX = RangeTenthInches * sin(DEGREES_TO_RADIANS*(StepAngleX) );// TODO! THIS SEEMS WRONG!!!
		TempY = RangeTenthInches * cos(DEGREES_TO_RADIANS*StepAngleY);// * cos(DEGREES_TO_RADIANS*(StepAngleX));
		TempZ = RangeTenthInches * sin(DEGREES_TO_RADIANS*StepAngleYZ * 1.1); // * cos(DEGREES_TO_RADIANS*(StepAngleX));

		// Now, translate origin from Kinect position to center of robot at floor.
		X = -TempX;	 // For Loki, Positive X is the Right side of the robot
		Y = TempY - DistanceFromFront;
		Z = TempZ + HeightAboveFloor;

		// Save the data in TenthInches
		//ROBOT_LOG( TRUE,  "Kinect Set: EnterCriticalSection g_csKinectPointCloudLock\n" )
		__itt_task_begin(pDomainKinectThread, __itt_null, __itt_null, psh_csKinectPointCloudLock);
		if( NULL == g_KinectPointCloud ) return; // TODO Kludge for shutdown crash.  Need better solution.
		EnterCriticalSection(&g_csKinectPointCloudLock);
			g_KinectPointCloud->Point3dArray[y][x].X = (int)X;
			g_KinectPointCloud->Point3dArray[y][x].Y = (int)Y;
			g_KinectPointCloud->Point3dArray[y][x].Z = (int)Z;
		LeaveCriticalSection(&g_csKinectPointCloudLock);
		__itt_task_end(pDomainKinectThread);
		//ROBOT_LOG( TRUE,  "Kinect Set: LeaveCriticalSection g_csKinectPointCloudLock\n" )
	}		

	/*
	if( m_DepthMousePointSelected )
	{
		if( (x == MouseX) && (y == MouseY) )
		{
			ROBOT_LOG( TRUE,  "MOUSE: Y=%d, X=%d:   Range=%4.1f, Y=%4.1f, X=%4.1f, Z=%4.1f\n", y, x,(RangeTenthInches/10.0), Y/10.0, X/10.0, Z/10.0  )
			m_DepthMousePointSelected = FALSE;
		}
	}
	*/
}

#endif // ( KINECT_SDK_TYPE_MICROSOFT == 1 )
// Everything in this file is for Microsoft SDK use only!
